/*
 * cwexp.cc
 *
 *  Created on: 15-Jan-2009
 *      Author: Leonard Tracy
 */


#include "ns3/core-module.h"
#include "ns3/common-module.h"
#include "ns3/uan-module.h"
#include "ns3/node-module.h"
#include "ns3/mobility-module.h"
#include "ns3/contrib-module.h"
#include "ns3/helper-module.h"

#include <iostream>
#include <fstream>

using namespace ns3;

NS_LOG_COMPONENT_DEFINE("CwExp");


class Experiment
{
public:

  Experiment() : m_BytesTotal(0),
                 m_asciitracefile("cwascii.tr") { }

  ///Runs an experiment with the current settings
  Gnuplot2dDataset Run(UanHelper &uan);
  void ReceivePacket(Ptr<Socket> socket);
  void UpdatePositions(NodeContainer &nodes, NetDeviceContainer &dev, uint32_t cw);
  void ResetData(std::ostream *os, uint32_t cw);
  uint32_t m_NumNodes;
  uint32_t m_DataRate;
  double m_Load;
  double m_Boundary;
  double m_Depth;
  uint32_t m_PacketSize;
  uint32_t m_BytesTotal;
  uint32_t m_CwMin;
  uint32_t m_CwMax;
  uint32_t m_Avgs;

  std::string m_mldatfile;
  std::string m_gnudatfile;
  std::string m_asciitracefile;

  Time m_SlotTime;


  Gnuplot2dDataset m_Data;

};

void
Experiment::ResetData(std::ostream *os, uint32_t cw)
{
  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << "  Resetting data");
  m_Data.Add(cw, m_BytesTotal);
  *os << cw << " " << m_BytesTotal << std::endl;
  m_BytesTotal = 0;
}

void
Experiment::UpdatePositions(NodeContainer &nodes, NetDeviceContainer &devices, uint32_t cw)
{
  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Updating positions");
  NodeContainer::Iterator it = nodes.Begin();
  for(;it != nodes.End();it++)
    {
      Ptr<MobilityModel> mp = (*it)->GetObject<MobilityModel>();
	  UniformVariable uv(0,m_Boundary);

      Vector a(uv.GetValue(),
               uv.GetValue(),
               70.0);
      NS_LOG_DEBUG("Setting new position: " << a);
    }

  Config::Set("/NodeList/*/DeviceList/*/Mac/CW", UintegerValue(cw));

}

void
Experiment::ReceivePacket(Ptr<Socket> socket)
{
  Ptr<Packet> packet;
  //NS_LOG_DEBUG("Receiving packet?");
  while(packet = socket->Recv())
    {
      //NS_LOG_DEBUG("Packetsize = " << packet->GetSize());
      m_BytesTotal += packet->GetSize();
      //std::cout << "Received packet: " << *packet;
    }
}
Gnuplot2dDataset
Experiment::Run(UanHelper &uan)
{
  uan.SetMac("ns3::UanMacCw", "CW", UintegerValue(m_CwMin), "SlotTime", TimeValue(m_SlotTime));

  NodeContainer nc = NodeContainer();
  NodeContainer sink = NodeContainer();
  nc.Create(m_NumNodes);
  sink.Create(1);

  PacketSocketHelper pktskth;
  pktskth.Install(nc);
  pktskth.Install(sink);

  Ptr<UanPropModelBh> prop = CreateObject<UanPropModelBh>("ConfigFile", StringValue("exbhconfig.cfg"));
  Ptr<UanPropModelIdeal> iprop = CreateObject<UanPropModelIdeal>();
  Ptr<UanChannel> channel = CreateObject<UanChannel>("PropagationModel", PointerValue(prop));

  //Create net device and nodes with UanHelper
  NetDeviceContainer devices = uan.Install(nc, channel);
  NetDeviceContainer sinkdev = uan.Install(sink, channel);

  MobilityHelper mobility;
  Ptr<ListPositionAllocator> pos = CreateObject<ListPositionAllocator>();
  pos->Add(Vector(m_Boundary/2, m_Boundary/2, m_Depth));
  UniformVariable urv(0,m_Boundary);

  for(uint32_t i=0;i<m_NumNodes;i++)
    {
      pos->Add(Vector(urv.GetValue(), urv.GetValue(), m_Depth));
    }

  mobility.SetPositionAllocator(pos);
  mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
  mobility.Install(sink);

  NS_LOG_DEBUG("Position of sink: " << sink.Get(0)->GetObject<MobilityModel>()->GetPosition());
  mobility.Install(nc);


  NodeContainer::Iterator it = nc.Begin();
  for(uint32_t i=0;i<nc.GetN();i++)
    {
      Ptr<Node> node = nc.Get(i);
      Ptr<NetDevice> dev = devices.Get(i);
      NS_LOG_DEBUG("Position of node " << dev->GetAddress() << ": " << node->GetObject<MobilityModel>()->GetPosition());
      it++;
    }

  PacketSocketAddress socket;
  socket.SetSingleDevice(sinkdev.Get(0)->GetIfIndex());
  socket.SetPhysicalAddress(sinkdev.Get(0)->GetAddress());
  socket.SetProtocol(0);
  NS_LOG_DEBUG("Remote address: " << socket);

  OnOffHelper app("ns3::PacketSocketFactory", Address(socket));
  app.SetAttribute("OnTime", RandomVariableValue(ConstantVariable ( (m_CwMax - m_CwMin + 1)*2500*m_Avgs)));
  app.SetAttribute("OffTime", RandomVariableValue(ConstantVariable(0)));
  app.SetAttribute("DataRate", DataRateValue(m_DataRate));
  app.SetAttribute("PacketSize", UintegerValue(m_PacketSize));

  //Schedule update events;
  Time simTime = Seconds(2000);

  ApplicationContainer apps = app.Install(nc);
  apps.Start(Seconds(0.5));
  Time now = Seconds(0.5);
  uint32_t cw = m_CwMin;

  std::ofstream mld(m_mldatfile.c_str());
  uint32_t an = 0;
  while(cw <= m_CwMax)
    {
      now += simTime;
      Simulator::Schedule(now, &Experiment::ResetData, this, &mld, cw);
      an++;
      if(an == m_Avgs)
        {
          cw++;
          an=0;
        }
      Simulator::Schedule(now, &Experiment::UpdatePositions, this, nc, devices, cw);
    }
  apps.Stop(now + simTime);

  Ptr<Node> sinkNode = sink.Get(0);
  TypeId psfid = TypeId::LookupByName("ns3::PacketSocketFactory");
  if(sinkNode->GetObject<SocketFactory>(psfid) == 0)
    {
      Ptr<PacketSocketFactory> psf = CreateObject<PacketSocketFactory>();
      sinkNode->AggregateObject(psf);
    }
  Ptr<Socket> sinkSocket = Socket::CreateSocket(sinkNode, psfid);
  sinkSocket->Bind(socket);
  sinkSocket->SetRecvCallback(MakeCallback(&Experiment::ReceivePacket, this));

  m_BytesTotal = 0;

  //GtkConfigStore config;
  //config.Configure();

//  std::ofstream ascii(m_asciitracefile.c_str());
//  uan.EnableAsciiAll(ascii);

  Simulator::Run();
  Simulator::Destroy();
  return m_Data;
}

int
main(int argc, char **argv)
{

  LogComponentEnable("CwExp", LOG_LEVEL_ALL);
  LogComponentEnable("UanMacCw", LOG_LEVEL_ALL);
  //LogComponentEnable("UanMacAloha", LOG_LEVEL_ALL);
  LogComponentEnable("UanPhyGen", LOG_LEVEL_ALL);

  //uan.SetMac("ns3::UanMacCw", "CW", UintegerValue(60), "SlotTime", TimeValue(Seconds(0.2)));

  Experiment exp;
  exp.m_NumNodes = 15;
  exp.m_Depth = 70;
  exp.m_Boundary = 500;
  exp.m_PacketSize = 32;
  exp.m_DataRate = 80;
  exp.m_CwMin = 2;
  exp.m_CwMax = 200;
  double slottime = 0.2;
  exp.m_Avgs = 7;
  exp.m_mldatfile = std::string("cwexpmlout.dat");
  std::string gnudatfile("cwexpgnuout.dat");
  std::string perModel = "ns3::UanPhyPerGenDefault";
  std::string sinrModel = "ns3::UanPhyCalcSinrDefault";

  CommandLine cmd;
  cmd.AddValue("NumNodes", "Number of transmitting nodes", exp.m_NumNodes);
  cmd.AddValue("Depth", "Depth of transmitting and sink nodes", exp.m_Depth);
  cmd.AddValue("RegionSize", "Size of boundary in meters", exp.m_Boundary);
  cmd.AddValue("PacketSize", "Generated packet size in bytes", exp.m_PacketSize);
  cmd.AddValue("DataRate", "DataRate in bps", exp.m_DataRate);
  cmd.AddValue("CwMin", "Min CW to simulate", exp.m_CwMin);
  cmd.AddValue("CwMax", "Max CW to simulate", exp.m_CwMax);
  cmd.AddValue("SlotTime", "Slot duration in seconds", slottime);
  cmd.AddValue("Averages", "Number of topologies to test for each cw point", exp.m_Avgs);
  cmd.AddValue("MlFile", "Name for plain text file output", exp.m_mldatfile);
  cmd.AddValue("GnuFile", "Name for GNU Plot output", gnudatfile);
  cmd.AddValue("PerModel", "PER model name", perModel);
  cmd.AddValue("SinrModel", "SINR model name", sinrModel);

  cmd.Parse(argc, argv);

  exp.m_SlotTime = Seconds(slottime);
  ObjectFactory obf;
  obf.SetTypeId(perModel);
  Ptr<UanPhyPer> per = obf.Create<UanPhyPer>();
  obf.SetTypeId(sinrModel);
  Ptr<UanPhyCalcSinr> sinr = obf.Create<UanPhyCalcSinr>();

  UanHelper uan;

  uan.SetPhy(std::string("ns3::UanPhyGen"),
             std::string("PerFunctor"), PointerValue(per),
             std::string("SinrFunctor"), PointerValue(sinr));


  Gnuplot gp;
  Gnuplot2dDataset ds;
  ds = exp.Run(uan);

  gp.AddDataset(ds);
  //std::ofstream of(filename.c_str());
  std::ofstream of(gnudatfile.c_str());
  gp.GenerateOutput(of);

}
