/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

#ifndef UANCHANNEL_H_
#define UANCHANNEL_H_

#include "ns3/net-device.h"
#include "ns3/channel.h"
#include "ns3/packet.h"
#include "ns3/uan-prop-model.h"
#include "ns3/uan-noise-model.h"

#include <list>
#include <vector>

namespace ns3
{

class UanNetDevice;
class UanPhy;
class UanTransducer;
class UanTxMode;


class UanChannel : public Channel
{
public:

  typedef std::vector<std::pair<Ptr<UanNetDevice>, Ptr<UanTransducer> > >
      UanDeviceList;

  UanChannel();
  UanChannel(std::string name);
  virtual ~UanChannel();

  static TypeId GetTypeId();

  //Methods inherrited from base class
  virtual uint32_t GetNDevices(void) const;
  virtual Ptr<NetDevice> GetDevice(uint32_t i) const;

  /**
   * \param src Transducer transmitting packet
   * \param packet Packet to be transmitted
   * \param txPowerDb Transmission power in dB
   *
   * Send a packet out on the channel
   */
  void TxPacket(Ptr<UanTransducer> src, Ptr<Packet> packet, double txPowerDb,
      UanTxMode txmode);

  /**
   * \param dev Net Device of node
   * \param trans Transducer of net device attached to this channel
   *
   * Adds device to receiver list for this channel
   */
  void AddDevice(Ptr<UanNetDevice> dev, Ptr<UanTransducer> trans);

  /**
   * \param prop Propagation model this channel will use for path loss/propagation delay
   */
  void SetPropagationModel(Ptr<UanPropagationModel> prop);

  /**
   * \param noise Noise model this channel will use to determine ambient channel noise.
   *
   */
  void SetNoiseModel(Ptr<UanNoiseModel> noise);

  /**
   * \params fKhz Frequency in kHz
   * \returns Ambient noise in dB/Hz on channel at a frequency
   */
  double GetNoiseDbHz(double fKhz);

private:

  UanDeviceList m_devList;
  Ptr<UanPropagationModel> m_prop;
  Ptr<UanNoiseModel> m_noise;

  void SendUp(uint32_t i, Ptr<Packet> packet, double rxPowerDb, UanTxMode txMode, UanPdp pdp);

};

}

#endif /*UANCHANNEL_H_*/
