/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

/**
 * \file uan-header-rc.h
 *
 * Headers (RTS, CTS, ACK, Data) used in UanMacRc and UanMacRcGw
 *
 */
#ifndef UANHEADERRC_H_
#define UANHEADERRC_H_

#include "ns3/header.h"
#include "ns3/nstime.h"

#include <set>

namespace ns3
{

/**
 * \class UanHeaderRcData
 *
 * \brief Extra data header information
 *
 * Adds prop. delay measure, and frame number info to
 * transmitted data packet
 */
class UanHeaderRcData : public ns3::Header
{
public:
  UanHeaderRcData();
  /**
   * \param frameNo Data frame # of reservation being transmitted
   * \param propDelay  Measured propagation delay found in handshaking
   * \note Prop. delay is transmitted with 16 bits and ms accuracy
   */
  UanHeaderRcData(uint8_t frameNum, Time propDelay);
  virtual ~UanHeaderRcData();

  static TypeId GetTypeId(void);

  /**
   * \param frameNum Data frame # of reservation being transmitted
   */
  void SetFrameNo(uint8_t frameNum);
  /**
   * \param propDelay Measured propagation delay found in handshaking
   * \note Prop. delay is transmitted with 16 bits and ms accuracy
   */
  void SetPropDelay(Time propDelay);
  /**
   * \returns Data frame # of reservation being transmitted
   */
  uint8_t GetFrameNo(void);
  /**
   * \returns Measured propagation delay found in handshaking
   * \note Prop. delay is transmitted with 16 bits and ms accuracy
   */
  Time GetPropDelay(void);

  //Inherrited methods
  virtual uint32_t GetSerializedSize (void) const;
  virtual void Serialize (Buffer::Iterator start) const;
  virtual uint32_t Deserialize (Buffer::Iterator start);
  virtual void Print (std::ostream &os) const;
  virtual TypeId GetInstanceTypeId(void) const;

private:
  uint8_t m_frameNo;
  Time m_propDelay;
};

/**
 * \class UanHeaderRcRts
 *
 * \brief RTS header
 *
 * Contains frame #, retry #, # frames, length, and timestamp
 */
class UanHeaderRcRts : public ns3::Header
{
public:
  UanHeaderRcRts();
  /**
   * \param frameNo Reservation frame #
   * \param retryNo Retry # of RTS packet
   * \param noFrames # of data frames in reservation
   * \param length # of bytes (including headers) in data
   * \param ts RTS TX timestamp
   * \note Timestamp is serialized into 32 bits with ms accuracy
   */
  UanHeaderRcRts(uint8_t frameNo, uint8_t retryNo, uint8_t noFrames, uint16_t length, Time ts);
  virtual ~UanHeaderRcRts();

  static TypeId GetTypeId(void);

  /**
   * \param fno TX frame #
   */
  void SetFrameNo(uint8_t fno);
  /**
   * \params no Number of data frames included in this reservation request
   */
  void SetNoFrames(uint8_t no);
  /**
   * \param timeStamp RTS transmission time
   */
  void SetTimeStamp(Time timeStamp);
  /**
   * \param length Total number of data bytes in reservation (including headers)
   * \note Timestamp is serialized with 32 bits in ms precision
   */
  void SetLength(uint16_t length);
  /**
   * \param no Retry number of this RTS (Used to match timestamp to correctly received RTS)
   */
  void SetRetryNo(uint8_t no);

  /**
   * \returns Frame #
   */
  uint8_t GetFrameNo(void);
  /**
   * \returns # of data frames in reservation
   */
  uint8_t GetNoFrames(void);
  /**
   * \returns TX time of the RTS packet
   * \note Timestamp is serialized with 32 bits in ms precision
   */
  Time GetTimeStamp(void);
  /**
   * \returns Total # of bytes in data packets for reservation (including headers)
   */
  uint16_t GetLength(void);
  /**
   * \returns Retry number of this RTS packet
   */
  uint8_t GetRetryNo(void);

  //Inherrited methods
  virtual uint32_t GetSerializedSize (void) const;
  virtual void Serialize (Buffer::Iterator start) const;
  virtual uint32_t Deserialize (Buffer::Iterator start);
  virtual void Print (std::ostream &os) const;
  virtual TypeId GetInstanceTypeId(void) const;

private:
  uint8_t m_frameNo;
  uint8_t m_noFrames;
  uint16_t m_length;
  Time m_timeStamp;
  uint8_t m_retryNo;
};

/**
 * \class UanHeaderRcCts
 *
 * \brief CTS header
 *
 * Includes RTS RX time, CTS TX time, delay until TX, RTS blocking period,
 * RTS tx period, rate #, and retry rate #
 */

class UanHeaderRcCts : public ns3::Header
{
public:
  UanHeaderRcCts();
  /**
   * \param frameNo Resrvation frame # being cleared
   * \param retryNo Retry # of received RTS packet
   * \param ts1 RX time of RTS packet at gateway
   * \param ts2 TX time of this CTS packet
   * \param delay Delay until transmission
   * \param rate Rate # for current cycle
   * \param rr Retry rate # for current cycle
   * \param blockTime Time duration to wait before beginning RTS transmissions
   * \param windowTime Time duration, following blockTime to allow RTS transmissions
   * \note Times are serialized, with ms precission, into 32 bit fields.
   */
  UanHeaderRcCts(uint8_t frameNo, uint8_t retryNo, Time ts1, Time ts2, Time delay,
                 uint16_t rate, uint16_t rr, Time blockTime, Time windowTime);
  virtual ~UanHeaderRcCts();

  static TypeId GetTypeId(void);

  /**
   * \param frameNo Frame # of RTS being cleared
   */
  void SetFrameNo(uint8_t frameNo);
  /**
   * \param timeStamp Time of RTS reception
   */
  void SetTimeStamp1(Time timeStamp);
  /**
   * \param timeStamp Time of CTS transmission
   */
  void SetTimeStamp2(Time timeStamp);
  /**
   * \param delay Time duration, from CTS TX, before first data frame arrival
   */
  void SetDelayToTx(Time delay);
  /**
   * \param no Retry number of RTS frame being cleared
   */
  void SetRetryNo(uint8_t no);
  /**
   * \param rate Rate number corresponding to data rate of current cycle
   */
  void SetRateNum(uint16_t rate);
  /**
   * \param rate Retry rate number for current cycle
   */
  void SetRetryRate(uint16_t rate);
  /**
   * \param t Time to block RTS transmissions
   */
  void SetBlockTime(Time t);
  /**
   * \param t Time duration following blocking time to allow RTS transmissions
   */
  void SetWindowTime(Time t);

  /**
   * \returns Frame # of RTS being cleared
   */
  uint8_t GetFrameNo(void);
  /**
   * \returns RX time of RTS being cleared
   */
  Time GetTimeStamp1(void);
  /**
   * \returns TX time of CTS packet
   */
  Time GetTimeStamp2(void);
  /**
   * \returns Delay from TX time of CTS packet until arrival of first data frame
   */
  Time GetDelayToTx(void);
  /**
   * \returns Retry # of RTS packet being cleared
   */
  uint8_t GetRetryNo(void);
  /**
   * \returns Rate # corresponding to data rate of current cycle
   */
  uint16_t GetRateNum(void);
  /**
   * \returns retry rate # of retry rate for current cycle
   */
  uint16_t GetRetryRate(void);
  /**
   * \returns Time duration to block RTS transmissions
   */
  Time GetBlockTime(void);
  /**
   * \returns Time duration after blocking time allowed for RTS transmissions
   */
  Time GetWindowTime(void);

  //Inherrited methods
  virtual uint32_t GetSerializedSize (void) const;
  virtual void Serialize (Buffer::Iterator start) const;
  virtual uint32_t Deserialize (Buffer::Iterator start);
  virtual void Print (std::ostream &os) const;
  virtual TypeId GetInstanceTypeId(void) const;

private:
  uint8_t m_frameNo;
  Time m_timeStamp1;
  Time m_timeStamp2;
  uint8_t m_retryNo;
  Time m_delay;
  Time m_blockTime;
  Time m_winTime;
  uint16_t m_retryRate;
  uint16_t m_rateNum;
};

class UanHeaderRcAck : public Header
{
public:
  UanHeaderRcAck();
  virtual ~UanHeaderRcAck();

  static TypeId GetTypeId(void);

  /**
   * \params frameNo Frame # of reservation being acknowledged
   */
  void SetFrameNo(uint8_t frameNo);
  /**
   * \params frame Data frame # being nacked
   */
  void AddNackedFrame(uint8_t frame);

  /**
   * \returns Set of nacked frames
   */
  const std::set<uint8_t> &GetNackedFrames(void) const;
  /**
   * \returns Reservation frame # being acknowledged.
   */
  uint8_t GetFrameNo(void) const;
  /**
   * \returns Number of data frames being NACKED
   */
  uint8_t GetNoNacks(void) const;

  //Inherrited methods
  virtual uint32_t GetSerializedSize (void) const;
  virtual void Serialize (Buffer::Iterator start) const;
  virtual uint32_t Deserialize (Buffer::Iterator start);
  virtual void Print (std::ostream &os) const;
  virtual TypeId GetInstanceTypeId(void) const;

private:
  uint8_t m_frameNo;
  std::set<uint8_t> m_nackedFrames;

};

}

#endif /* UANHEADERRC_H_ */
