/*
 * uan-mac-rc-gw.cc
 *
 *  Created on: 5-Mar-2009
 *      Author: ltracy
 */

#include "uan-mac-rc-gw.h"
#include "uan-mac-rc.h"
#include "uan-header-common.h"
#include "uan-header-rc.h"
#include "uan-phy.h"
#include "uan-tx-mode.h"

#include "ns3/assert.h"
#include "ns3/log.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/nstime.h"
#include "ns3/double.h"
#include "ns3/uinteger.h"

#include <utility>
#include <set>
#include <map>

NS_LOG_COMPONENT_DEFINE("UanMacRcGw");

namespace ns3
{

NS_OBJECT_ENSURE_REGISTERED(UanMacRcGw);

bool
operator < (UanAddress &a, UanAddress &b)
{
  return a.GetAsInt() < b.GetAsInt();
}


UanMacRcGw::UanMacRcGw() :
  m_state(IDLE),
  m_currentRate(0)
{
  UanHeaderCommon ch;
  UanHeaderRcRts rts;
  UanHeaderRcCts cts;
  UanHeaderRcAck ack;

  m_rtsSize = ch.GetSerializedSize() + rts.GetSerializedSize();
  m_ctsSize = ch.GetSerializedSize() + cts.GetSerializedSize();
  m_ackSize = ch.GetSerializedSize() + ack.GetSerializedSize();

  NS_LOG_DEBUG("Gateway initialized");
}

UanMacRcGw::~UanMacRcGw()
{
  // TODO Auto-generated destructor stub
}

TypeId
UanMacRcGw::GetTypeId(void)
{
  TypeId tid = TypeId("ns3::UanMacRcGw")
      .SetParent<UanMac>()
      .AddConstructor<UanMacRcGw>()
      .AddAttribute("MaxReservations",
                    "Maximum number of reservations to accept per cycle",
                    UintegerValue(10),
                    MakeUintegerAccessor(&UanMacRcGw::m_maxRes),
                    MakeUintegerChecker<uint32_t>())
      .AddAttribute("NumberOfRates",
                    "Number of rates per Phy layer",
                    UintegerValue(1023),
                    MakeUintegerAccessor(&UanMacRcGw::m_numRates),
                    MakeUintegerChecker<uint32_t>())
      .AddAttribute("RetryRate",
                    "Number of retry rates per second at non-gateway nodes",
                    DoubleValue(1/10.0),
                    MakeDoubleAccessor(&UanMacRcGw::m_retryRate),
                    MakeDoubleChecker<double>())
      .AddAttribute("MaxPropDelay",
                    "Maximum propagation delay between gateway and non-gateway nodes",
                    TimeValue(Seconds(2)),
                    MakeTimeAccessor(&UanMacRcGw::m_maxDelta),
                    MakeTimeChecker())
      .AddAttribute("SIFS",
                    "Spacing between frames to account for timing error and processing delay",
                    TimeValue(Seconds(0.2)),
                    MakeTimeAccessor(&UanMacRcGw::m_sifs),
                    MakeTimeChecker())
      .AddAttribute("NumberOfNodes",
                    "Number of non-gateway nodes in this gateway's neighborhood",
                    UintegerValue(10),
                    MakeUintegerAccessor(&UanMacRcGw::m_numNodes),
                    MakeUintegerChecker<uint32_t>())
      .AddAttribute("MinRetryRate",
                    "Smallest allowed RTS retry rate",
                    DoubleValue(0.01),
                    MakeDoubleAccessor(&UanMacRcGw::m_minRetryRate),
                    MakeDoubleChecker<double>())
      .AddAttribute("RetryStep",
                    "Retry rate increment",
                    DoubleValue(0.01),
                    MakeDoubleAccessor(&UanMacRcGw::m_retryStep),
                    MakeDoubleChecker<double>())
      .AddAttribute("NumberOfRetryRates",
                    "Number of retry rates",
                    UintegerValue(100),
                    MakeUintegerAccessor(&UanMacRcGw::m_numRetryRates),
                    MakeUintegerChecker<uint16_t>())
      .AddAttribute("TotalRate",
                    "Total available channel rate in bps (for a single channel, without splitting reservation channel)",
                    UintegerValue(4096),
                    MakeUintegerAccessor(&UanMacRcGw::m_totalRate),
                    MakeUintegerChecker<uint32_t>())
      .AddAttribute("RateStep",
                    "Increments available for rate assignment in bps",
                    UintegerValue(4),
                    MakeUintegerAccessor(&UanMacRcGw::m_rateStep),
                    MakeUintegerChecker<uint32_t>())

//      .AddTraceSource("Enqueue",
//                    "A (data) packet arrived at MAC for transmission",
//                     MakeTraceSourceAccessor(&UanMacRcGw::m_enqueueLogger))
//      .AddTraceSource("Dequeue",
//                    "A (data) packet was passed down to PHY from MAC",
//                    MakeTraceSourceAccessor(&UanMacRcGw::m_dequeueLogger))
      .AddTraceSource("RX",
                      "A packet was destined for and received at this MAC layer",
                      MakeTraceSourceAccessor(&UanMacRcGw::m_rxLogger))
      .AddTraceSource("Cycle",
                      "Trace cycle statistics",
                      MakeTraceSourceAccessor(&UanMacRcGw::m_cycleLogger))

      ;

  return tid;
}

Address
UanMacRcGw::GetAddress(void)
{
  return m_address;
}

void
UanMacRcGw::SetAddress(UanAddress addr)
{
  m_address = addr;
}

bool
UanMacRcGw::Enqueue(Ptr<Packet> packet, const Address &dest, uint16_t protocolNumber)
{
  NS_LOG_WARN("RCMAC Gateway transmission to acoustic nodes is not yet implemented");
  return false;
}

void
UanMacRcGw::SetForwardUpCb(Callback<void, Ptr<Packet>, const UanAddress&> cb)
{
  //NS_LOG_WARN("RCMAC Gateway transmission to acoustic nodes is not yet implemented");
  m_forwardUpCb = cb;
}

void
UanMacRcGw::AttachPhy(Ptr<UanPhy> phy)
{
  m_phy = phy;
  phy->SetReceiveOkCallback(MakeCallback(&UanMacRcGw::ReceivePacket, this));
  phy->SetReceiveErrorCallback(MakeCallback(&UanMacRcGw::ReceiveError, this));
}

void
UanMacRcGw::ReceiveError(Ptr<Packet> pkt, double sinr)
{
//  std::string type;
//  UanHeaderCommon ch;
//  pkt->PeekHeader(ch);
//  switch(ch.GetType())
//  {
//    case UanMacRc::TYPE_DATA:
//      type = "DATA";
//      break;
//    case UanMacRc::TYPE_RTS:
//      type= "RTS";
//      break;
//    case UanMacRc::TYPE_CTS:
//      type="CTS";
//      break;
//    case UanMacRc::TYPE_ACK:
//      type="ACK";
//      break;
//    case UanMacRc::TYPE_GWPING:
//      type="GWPING";
//      break;
//    default:
//      type="UNKNOWN";
//      break;
//  }
//  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW Received in error: Packet of type " << type << " from " << ch.GetSrc() << " with SINR " << sinr << " and mode ?");
}
Address
UanMacRcGw::GetBroadcast(void) const
{
  return UanAddress::GetBroadcast();
}

void
UanMacRcGw::ReceivePacket(Ptr<Packet> pkt, double sinr, UanTxMode mode)
{
  UanHeaderCommon ch;
  pkt->PeekHeader(ch);
  //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW Received packet from " << ch.GetSrc());

  if(ch.GetDest() == m_address || ch.GetDest() == UanAddress::GetBroadcast())
    {
      m_rxLogger(pkt, mode);
    }
  else
    return;

  //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW Received packet from " << ch.GetSrc());
  pkt->RemoveHeader(ch);

  switch(ch.GetType())
  {
    case UanMacRc::TYPE_DATA:
      {
      UanHeaderRcData dh;
      pkt->RemoveHeader(dh);
      //NS_LOG_DEBUG(Simulator::Now() << "GW Adding prop delay for " << ch.GetSrc() << " of: " << dh.GetPropDelay());
      m_propDelay[ch.GetSrc()] = dh.GetPropDelay();
      if(m_ackData.find(ch.GetSrc()) == m_ackData.end())
        {
          NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GATEWAY Received unexpected data packet");
        }
      else
        {
          NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW Received data packet from " << ch.GetSrc() << " length = " << pkt->GetSize());
          m_ackData[ch.GetSrc()].rxFrames.insert(dh.GetFrameNo());
        }
      m_forwardUpCb(pkt, ch.GetSrc());
      }
      break;
    case UanMacRc::TYPE_GWPING:
    case UanMacRc::TYPE_RTS:
      if(m_state == CTSING)
        return;

      {
        UanHeaderRcRts rh;
        pkt->RemoveHeader(rh);

        if(m_requests.find(ch.GetSrc()) == m_requests.end())
          {
            Request req;
            req.numFrames = rh.GetNoFrames();
            req.rxTime = Simulator::Now();
            req.frameNo = rh.GetFrameNo();
            req.retryNo = rh.GetRetryNo();
            req.length = rh.GetLength();
            NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW storing reservation from " << ch.GetSrc() << " with length " << req.length);
            m_requests.insert(std::make_pair(ch.GetSrc(), req));
            std::map<UanAddress, Time>::iterator it = m_propDelay.find(ch.GetSrc());
            if(it == m_propDelay.end())
              m_sortedRes.insert(std::make_pair(m_maxDelta, ch.GetSrc()));
            else
              m_sortedRes.insert(std::make_pair((*it).second, ch.GetSrc()));
          }
        else
          {
            //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW received duplicate RTS from " << ch.GetSrc());
          }

      }
      if(m_state == IDLE)
        StartCycle();
      break;
    case UanMacRc::TYPE_CTS:
      NS_FATAL_ERROR("Received CTS at GW.  Currently only support single GW network!");
      break;
    case UanMacRc::TYPE_ACK:
      NS_FATAL_ERROR("Received ACK at GW.  Currently only support single GW network!");
      break;
    default:
      NS_FATAL_ERROR("Received unknown packet at GW!");
  }
}

void
UanMacRcGw::StartCycle(void)
{


  uint32_t numRts = m_sortedRes.size();

  if(numRts)
    NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Simulator starting non-empty cycle");
  else
    NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Simulator starting EMPTY cycle");



  //Calculate dataRate
  uint32_t totalBytes=0;
  uint32_t totalFrames=0;

  if(numRts > 0)
    {
      std::map<UanAddress, Request>::iterator rit = m_requests.begin();
      for(;rit != m_requests.end(); rit++)
        {
          totalBytes += (*rit).second.length;
          totalFrames += (*rit).second.numFrames;
        }
    }

  //Linear search through variables algorithm (Maybe bug in this)
//  int32_t neededArrivals=0;
//  if(m_numNodes > m_maxRes)
//    neededArrivals =  (log(m_numNodes - m_maxRes) - log(m_numNodes)) / log( 1.0 - 1.0 / (double) m_numNodes);
//
//  if(m_numNodes <= m_maxRes || neededArrivals > 2.0*m_numNodes)
//    neededArrivals = 2*m_numNodes;

  //NS_LOG_DEBUG("Calculated needed arrivals = " << neededArrivals << " for " << m_numNodes << " nodes and " << m_maxRes << " max reservations");

  //Just using constant a
  //int32_t neededArrivals = m_maxRes;

//  double optRateError = 0;
//  uint32_t optRateNum = 0;
//  uint32_t optRateValue =  m_phy->GetMode(m_numRates).GetDataRateBps();
//  uint16_t optRetryRate = 0;
//  double optWinSize;

//  for(uint16_t rr = 0; rr < m_numRetryRates; rr++)
//    {
//      uint32_t semiOptRateNum = m_numRates;
//      double semiOptRateError = 1e300;
//      double semiOptRateValue = m_phy->GetMode(m_numRates).GetDataRateBps();
//      double semiOptWinSize = 0;
//      for(uint32_t i=0;i<=m_numRates-1;i++)
//        {
//          double rate = m_phy->GetMode(i).GetDataRateBps();
//          double ctlrate = m_phy->GetMode(i+m_numRates).GetDataRateBps();
//          double winSize;
//          if(numRts > 0)
//            {
//              winSize = (double) (totalBytes)*8.0/rate + 2*m_maxDelta.GetSeconds() / (numRts + 1.0) / 2.0 + m_sifs.GetSeconds()*totalFrames - m_rtsSize*8/ctlrate;
//
//            }
//          else
//            winSize = m_maxDelta.GetSeconds() / 2.0 - m_rtsSize*8/ctlrate;
//          if(winSize < 0)
//             winSize = 0.0;
//
//          double expReqs = winSize*m_numNodes*(m_minRetryRate+rr*m_retryStep)*std::exp(-2.0 *m_numNodes * (m_minRetryRate+rr*m_retryStep)*m_rtsSize * 8 / ctlrate);
//
//          double rateError = std::abs((double) neededArrivals - expReqs);
//          //NS_LOG_DEBUG("With i = " << i << " expReqs = " << expReqs << " rateError = " << rateError << " rr = " << rr);
//          if( (rateError > semiOptRateError && i != 0) || (i == m_numRates))
//            break;
//          else
//            {
//              semiOptRateError = rateError;
//              semiOptRateNum = i;
//              semiOptRateValue = ctlrate;
//              semiOptWinSize = winSize;
//            }
//        }
//      if(semiOptRateNum <= optRateNum || rr==0 || optRateError - semiOptRateError >= 1.0)
//        {
//          //NS_LOG_DEBUG("For retry rate = " << (m_minRetryRate+rr*m_retryStep) << " found opt rate num " << semiOptRateNum << " for ctrl rate " << semiOptRateValue);
//          optRateNum = semiOptRateNum;
//          optRateValue = semiOptRateValue;
//          optRetryRate = rr;
//          optRateError = semiOptRateError;
//          optWinSize = semiOptWinSize;
//        }
//    }


  //New algorithm

  double minRate = m_phy->GetMode(m_numRates).GetDataRateBps();

  //Temp values
  double beta = numRts*m_sifs.GetSeconds() + 2.0*m_maxDelta.GetSeconds()/(numRts+1.0);
  double aep = 0.5 + m_maxRes * exp(1.0);
  double c = totalBytes*8.0;

  double optalpha = aep / ( (1/(2.0*m_rtsSize*8.0))*(c + beta * m_totalRate) + aep  );

  double optctlrate = m_totalRate*optalpha;

  double temprate = (optctlrate - minRate) / ((double) m_rateStep) + 0.5;
  m_currentRate =  (uint32_t) temprate;
  NS_LOG_DEBUG("Found alpha: " << optalpha << " Found associated rate = " << optctlrate << " Giving rate number: " << temprate);
  double optret = optalpha*m_totalRate / (2.0*m_numNodes*m_rtsSize*8.0);

  double dataRate = m_phy->GetMode(m_currentRate).GetDataRateBps();


  if(optret < m_minRetryRate)
    {
      NS_LOG_WARN("Gateway found optimum RTS retry rate is below minimum");
      m_currentRetryRate = 0;
    }
  else
    {
      m_currentRetryRate = (uint16_t)  ((optret - m_minRetryRate) / m_retryStep + 0.5) ;
    }

  double actualrr = m_currentRetryRate * m_retryStep + m_minRetryRate;

  uint32_t optRateValue =  m_phy->GetMode(m_currentRate+m_numRates).GetDataRateBps();

  double optWinSize = (double) (totalBytes)*8.0/dataRate + 2*m_maxDelta.GetSeconds() / (numRts + 1.0) + m_sifs.GetSeconds()*totalFrames - m_rtsSize*8/optRateValue;
  double optWinSizeT = (double) (totalBytes)*8.0/(m_totalRate*(1.0 - optalpha)) + 2.0*m_maxDelta.GetSeconds() / (numRts + 1.0) + m_sifs.GetSeconds()*totalFrames - m_rtsSize*8/(optctlrate);



  double cycleSeconds = totalBytes*8 / dataRate + totalFrames * 3*m_sifs.GetSeconds() + (m_ctsSize + 5)*8*totalFrames/dataRate + m_maxDelta.GetSeconds();
  double expReqs = optWinSize*m_numNodes*(actualrr)*std::exp(-2.0 *m_numNodes * (actualrr)*m_rtsSize * 8 / optRateValue);
  double optexpreqs = optWinSizeT * m_numNodes*(optret) * std::exp(-2.0 * m_numNodes * (optret)*m_rtsSize * 8 / (optctlrate));
  double optRateError = std::abs(expReqs - m_maxRes);

  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW ALGORITHM NEW INFO:  expreqs " << expReqs << " With theoretical value: " << optexpreqs << " optalpha " << optalpha << " optret " << optret << " actualrr " << actualrr);
  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW ALGORITHM: For " << numRts << " reservations (totBytes = " << totalBytes << ", " << optWinSize << "), chose rate #=" << m_currentRate << " for ctlrate=" << optRateValue << " with retry rate value " << m_retryStep*m_currentRetryRate+m_minRetryRate << " and rate error " << optRateError << " total cycle time = " << cycleSeconds);

  Time ctsTxTime = Seconds(m_ctsSize*8/dataRate);
  if(numRts == 0)
    {
      UanHeaderRcCts ctsh;
      ctsh.SetBlockTime(m_sifs);
      ctsh.SetWindowTime(Seconds(optWinSize));
      ctsh.SetRateNum(m_currentRate);
      ctsh.SetRetryRate(m_currentRetryRate);

      UanHeaderCommon ch(m_address, UanAddress::GetBroadcast(), UanMacRc::TYPE_CTS);
      Ptr<Packet> p = Create<Packet>();
      p->AddHeader(ctsh);
      p->AddHeader(ch);
      SendPacket(p, m_currentRate);
      Simulator::Schedule(ctsTxTime+Seconds(optWinSize) + m_sifs + Seconds(m_rtsSize*8/m_phy->GetMode(m_currentRate + m_numRates).GetDataRateBps()), &UanMacRcGw::StartCycle, this);
      m_state = INCYCLE;
      return;
    }




  Time nextEarliest = Seconds(numRts*m_ctsSize*8/dataRate)  + Seconds(numRts*m_sifs.GetSeconds()) + m_sifs;

  m_state = CTSING;
  Simulator::Schedule(nextEarliest, &UanMacRcGw::CycleStarted, this);
  Time endCts = nextEarliest;
  Time nextCts = nextEarliest - m_sifs - ctsTxTime;
  std::set<std::pair<Time, UanAddress> >::iterator it = m_sortedRes.begin();
  Time minPdelay = (*it).first;
  for(; it != m_sortedRes.end(); it++)
    {
      Request req = m_requests[(*it).second];
      Time pdelay = (*it).first;

      NS_LOG_DEBUG(Simulator::Now() << " GW: Scheduling request for prop. delay " << pdelay.GetSeconds() << " for " << (*it).second);

      AckData newData;
      newData.expFrames = req.numFrames;
      newData.frameNo = req.frameNo;

      UanAddress dest = (*it).second;

      m_ackData.insert(std::make_pair(dest, newData));

      Ptr<Packet> cts = Create<Packet>();
      UanHeaderRcCts ctsh;

      ctsh.SetTimeStamp1(req.rxTime);
      ctsh.SetTimeStamp2(nextCts + Simulator::Now());

      ctsh.SetRateNum(m_currentRate);
      ctsh.SetFrameNo(req.frameNo);
      ctsh.SetRetryNo(req.retryNo);
      ctsh.SetRetryRate(m_currentRetryRate);
      ctsh.SetBlockTime(endCts - nextCts - ctsTxTime);
      ctsh.SetWindowTime(Seconds(optWinSize));


      Time earliestArr = nextCts + pdelay + pdelay + ctsTxTime + m_sifs;
      Time arrivalTime = std::max(earliestArr, nextEarliest);


      ctsh.SetDelayToTx(arrivalTime - nextCts);
      NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW Scheduling reception for " << (uint32_t) req.numFrames << " frames " << (Simulator::Now() + arrivalTime).GetSeconds() << "  (delaytiltx of " << (arrivalTime-nextCts).GetSeconds() << ")  Total length is " << req.length << " with txtime " << req.length*8/dataRate << " seconds");
      //NS_LOG_DEBUG(" GW Stamped CTS with tx time " << (nextCts + Simulator::Now()).GetSeconds());

      UanHeaderCommon ch;
      ch.SetDest(dest);
      ch.SetSrc(m_address);

      ch.SetType(UanMacRc::TYPE_CTS);

      cts->AddHeader(ctsh);
      cts->AddHeader(ch);

      Simulator::Schedule(nextCts, &UanMacRcGw::SendPacket, this, cts, m_currentRate);
      nextCts = nextCts - m_sifs - ctsTxTime;
      nextEarliest = arrivalTime + Seconds(req.length*8/dataRate) + Seconds(m_sifs.GetSeconds()*req.numFrames);
    }
  m_requests.clear();
  m_sortedRes.clear();
  Simulator::Schedule(nextEarliest, &UanMacRcGw::EndCycle, this);


  m_cycleLogger(Simulator::Now(), minPdelay, numRts, totalBytes, optWinSize, optRateValue, actualrr);
}

void
UanMacRcGw::CycleStarted()
{
  m_state = INCYCLE;
}
void
UanMacRcGw::EndCycle()
{

  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW Ending cycle");

  Time nextAck = Seconds(0);

  Time ackTime = Seconds(m_ackSize * 8.0/ m_phy->GetMode(m_currentRate).GetDataRateBps());

  std::map<UanAddress, AckData>::iterator it = m_ackData.begin();
  for(; it != m_ackData.end(); it++)
    {
      UanAddress dest = (*it).first;
      AckData &data = (*it).second;

      std::list<uint32_t> toNack;
      for(uint32_t i=0;i<data.expFrames;i++)
        {
          if(data.rxFrames.find(i) == data.rxFrames.end())
            toNack.push_back(i);
        }
      UanHeaderCommon ch;
      ch.SetDest(dest);
      ch.SetSrc(m_address);
      ch.SetType(UanMacRc::TYPE_ACK);
      UanHeaderRcAck ah;
      ah.SetFrameNo(data.frameNo);
      std::list<uint32_t>::iterator nit = toNack.begin();
      for(; nit != toNack.end(); nit++)
        {
          ah.AddNackedFrame(*nit);
        }

      Ptr<Packet> ack = Create<Packet>();
      ack->AddHeader(ah);
      ack->AddHeader(ch);
      Simulator::Schedule(nextAck, &UanMacRcGw::SendPacket, this, ack, m_currentRate);
      nextAck = nextAck + ackTime + m_sifs;
    }
  m_ackData.clear();
  Simulator::Schedule(nextAck, &UanMacRcGw::StartCycle, this);

}
void
UanMacRcGw::SendPacket(Ptr<Packet> pkt, uint32_t rate)
{
  UanHeaderCommon ch;
  pkt->PeekHeader(ch);
  std::string type;
  switch(ch.GetType())
  {
    case UanMacRc::TYPE_DATA:
      type = "DATA";
      break;
    case UanMacRc::TYPE_RTS:
      type= "RTS";
      break;
    case UanMacRc::TYPE_CTS:
      type="CTS";
      break;
    case UanMacRc::TYPE_ACK:
      type="ACK";
      break;
    case UanMacRc::TYPE_GWPING:
      type="GWPING";
      break;
    default:
      type="UNKNOWN";
      break;
  }
  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " GW sending " << type << " packet with size " << pkt->GetSize() << " to " << ch.GetDest() << " at rate " << rate);
  m_phy->SendPacket(pkt, rate);
}

}// namespace ns3
