/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

#ifndef UANMACRCGW_H_
#define UANMACRCGW_H_

#include "uan-mac.h"
#include "uan-address.h"

#include "ns3/nstime.h"
#include "ns3/traced-callback.h"




#include <set>
#include <map>

namespace ns3
{

class UanTxMode;

/**
 * \class UanMacRcGw
 *
 * MAC protocol for gateway nodes in a network using reservation channel MAC (UanMacRc)
 */
class UanMacRcGw: public ns3::UanMac
{
public:
  UanMacRcGw();
  virtual ~UanMacRcGw();

  static TypeId GetTypeId(void);

  //Inherited virtual functions
  virtual Address GetAddress(void);
  virtual void SetAddress(UanAddress addr);
  virtual bool Enqueue(Ptr<Packet> packet, const Address &dest, uint16_t protocolNumber);
  virtual void SetForwardUpCb(Callback<void, Ptr<Packet>, const UanAddress&> cb);
  virtual void AttachPhy(Ptr<UanPhy> phy);
  virtual Address GetBroadcast(void) const;

private:
  enum State {IDLE, INCYCLE, CTSING};
  State m_state;

  struct Request
  {
    uint8_t numFrames;
    uint8_t frameNo;
    uint8_t retryNo;
    uint16_t length;
    Time rxTime;
  };

  struct AckData
  {
    uint8_t frameNo;
    std::set<uint8_t> rxFrames;
    uint8_t expFrames;
  };
  Callback<void, Ptr<Packet>, const UanAddress& > m_forwardUpCb;
  Ptr<UanPhy> m_phy;
  UanAddress m_address;
  Time m_maxDelta;
  Time m_sifs;
  uint32_t m_maxRes;
  uint32_t m_numRates;
  uint32_t m_rtsSize;
  uint32_t m_ctsSize;
  uint32_t m_ackSize;
  double m_retryRate;
  uint16_t m_currentRetryRate;
  uint32_t m_currentRate;
  uint32_t m_numNodes;
  uint32_t m_totalRate;
  uint32_t m_rateStep;

  uint16_t m_numRetryRates;
  double m_minRetryRate;
  double m_retryStep;

  std::map<UanAddress, Time> m_propDelay;

  std::map<UanAddress, AckData> m_ackData;


  std::map<UanAddress, Request> m_requests;
  std::set<std::pair<Time, UanAddress> > m_sortedRes;

  TracedCallback<Ptr<const Packet>, UanTxMode > m_rxLogger;

  //Start time, min p-delay, reservations, frames, bytes, window size, ctl rate, retry rate
  TracedCallback<Time, Time, uint32_t, uint32_t, double, uint32_t, double> m_cycleLogger;
//  TracedCallback<Ptr<const Packet>, uint16_t > m_enqueueLogger;
//  TracedCallback<Ptr<const Packet>, uint16_t > m_dequeueLogger;

  void ReceivePacket(Ptr<Packet> pkt, double sinr, UanTxMode mode);
  void StartCycle(void);
  void EndCycle(void);
  void SendPacket(Ptr<Packet> pkt, uint32_t rate);
  void CycleStarted(void);
  void ReceiveError(Ptr<Packet> pkt, double sinr);
};

/**
 * Defined for use in UanMacRcGw
 */
bool operator < (UanAddress &a, UanAddress &b);

}

#endif /* UANMACRCGW_H_ */
