/*
 * uan-mac-rc.cc
 *
 *  Created on: 1-Mar-2009
 *      Author: ltracy
 */

#include "uan-mac-rc.h"
#include "uan-header-rc.h"
#include "uan-tx-mode.h"
#include "uan-phy.h"
#include "uan-header-common.h"
#include "uan-phy-dual.h"

#include "ns3/log.h"
#include "ns3/nstime.h"
#include "ns3/simulator.h"
#include "ns3/random-variable.h"
#include "ns3/assert.h"
#include "ns3/double.h"
#include "ns3/uinteger.h"

#include <list>
#include <utility>


NS_LOG_COMPONENT_DEFINE("UanMacRc");
namespace ns3
{

  uint32_t m_cntrlSends=0;

NS_OBJECT_ENSURE_REGISTERED(UanMacRc);

UanMacRc::UanMacRc() :
  m_state(UNASSOCIATED),
  m_rtsBlocked(false),
  m_currentRate(10),
  m_frameNo(0)
{
  UanHeaderCommon ch;
  UanHeaderRcCts ctsh;

  m_ctsSize = ch.GetSerializedSize() + ctsh.GetSerializedSize();


}

UanMacRc::~UanMacRc()
{
  // TODO Auto-generated destructor stub
}

TypeId
UanMacRc::GetTypeId(void)
{
  TypeId tid = TypeId("ns3::UanMacRc")
      .SetParent<UanMac>()
      .AddConstructor<UanMacRc>()
      .AddAttribute("RetryRate",
                    "Number of retry attempts per second (of RTS/GWPING)",
                    DoubleValue(1/5.0),
                    MakeDoubleAccessor(&UanMacRc::m_retryRate),
                    MakeDoubleChecker<double>())
      .AddAttribute("MaxFrames",
                    "Maximum number of frames to include in a single RTS",
                    UintegerValue(1),
                    MakeUintegerAccessor(&UanMacRc::m_maxFrames),
                    MakeUintegerChecker<uint32_t>())
      .AddAttribute("QueueLimit",
                    "Maximum packets to queue at MAC",
                    UintegerValue(10),
                    MakeUintegerAccessor(&UanMacRc::m_queueLimit),
                    MakeUintegerChecker<uint32_t>())
      .AddAttribute("SIFS",
                    "Spacing to give between frames (this should match gateway)",
                    TimeValue(Seconds(0.2)),
                    MakeTimeAccessor(&UanMacRc::m_sifs),
                    MakeTimeChecker())
      .AddAttribute("NumberOfRates",
                    "Number of rate divisions supported by each PHY",
                    UintegerValue(0),
                    MakeUintegerAccessor(&UanMacRc::m_numRates),
                    MakeUintegerChecker<uint32_t>())
      .AddAttribute("MinRetryRate",
                    "Smallest allowed RTS retry rate",
                    DoubleValue(0.01),
                    MakeDoubleAccessor(&UanMacRc::m_minRetryRate),
                    MakeDoubleChecker<double>())
      .AddAttribute("RetryStep",
                    "Retry rate increment",
                    DoubleValue(0.01),
                    MakeDoubleAccessor(&UanMacRc::m_retryStep),
                    MakeDoubleChecker<double>())
      .AddAttribute("NumberOfRetryRates",
                    "Number of retry rates",
                    UintegerValue(100),
                    MakeUintegerAccessor(&UanMacRc::m_numRetryRates),
                    MakeUintegerChecker<uint16_t>())
      .AddTraceSource("Enqueue",
                      "A (data) packet arrived at MAC for transmission",
                      MakeTraceSourceAccessor(&UanMacRc::m_enqueueLogger))
      .AddTraceSource("Dequeue",
                      "A (data) packet was passed down to PHY from MAC",
                      MakeTraceSourceAccessor(&UanMacRc::m_dequeueLogger))
      .AddTraceSource("RX",
                      "A packet was destined for and received at this MAC layer",
                      MakeTraceSourceAccessor(&UanMacRc::m_rxLogger))

      ;
  return tid;
}


Address
UanMacRc::GetAddress(void)
{
  return m_address;
}
void
UanMacRc::SetAddress(UanAddress addr)
{
  m_address = addr;
}
bool
UanMacRc::Enqueue(Ptr<Packet> packet, const Address &dest, uint16_t protocolNumber)
{
  if(protocolNumber > 0)
    NS_LOG_WARN("Warning: UanMacRc does not support multiple protocols.  protocolNumber argument to Enqueue is being ignored");


  if(m_pktQueue.size() >= m_queueLimit)
    return false;

  m_pktQueue.push_back(std::make_pair(packet, UanAddress::ConvertFrom(dest)));

  switch(m_state)
  {
    case UNASSOCIATED:
      Associate();
      return true;
    case IDLE:
      if(!m_rtsEvent.IsRunning())
        SendRts();
      return true;
    case GWPSENT:
    case RTSSENT:
    case DATATX:
      return true;
  }

  return true;
}
void
UanMacRc::SetForwardUpCb(Callback<void, Ptr<Packet>, const UanAddress&> cb)
{
  m_forwardUpCb = cb;
}
void
UanMacRc::AttachPhy(Ptr<UanPhy> phy)
{
  m_phy = phy;
  m_phy->SetReceiveOkCallback(MakeCallback(&UanMacRc::ReceiveOkFromPhy, this));
}

Address
UanMacRc::GetBroadcast(void) const
{
  return UanAddress::GetBroadcast();
}

void
UanMacRc::ReceiveOkFromPhy(Ptr<Packet> pkt, double sinr, UanTxMode mode)
{


  //Filter and log packets with headers intact
  UanHeaderCommon ch;
  pkt->RemoveHeader(ch);
  if(ch.GetDest() == m_address || ch.GetDest() == UanAddress::GetBroadcast())
    m_rxLogger(pkt, mode);

  switch(ch.GetType())
  {
    case TYPE_DATA:

      if(ch.GetDest() == m_address)
        {
          NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " UanMacRc Receiving DATA packet from PHY");
          UanHeaderRcData dh;
          pkt->RemoveHeader(dh);
          m_forwardUpCb(pkt, ch.GetSrc());
        }
      break;
    case TYPE_RTS:
      //Currently don't respond to RTS packets at non-gateway nodes
      break;
    case TYPE_CTS:

      {
      m_assocAddr = ch.GetSrc();
      m_rtsBlocked = true;
      UanHeaderRcCts ctsh;
      pkt->PeekHeader(ctsh);
      m_currentRate = ctsh.GetRateNum();
      m_retryRate = m_minRetryRate + m_retryStep*ctsh.GetRetryRate();
      if(!m_startAgain.IsRunning())
        {
          //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " Scheduling start again with delay " << ctsh.GetBlockTime().GetSeconds());
          m_startAgain = Simulator::Schedule(ctsh.GetBlockTime(), &UanMacRc::StartRtsing, this);
          Simulator::Schedule(ctsh.GetBlockTime() + ctsh.GetWindowTime(), &UanMacRc::BlockRtsing, this);
        }

      }
      if(ch.GetDest() != m_address)
        return;
      //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " UanMacRc Receiving CTS packet from PHY");

      if(m_state == GWPSENT)
        {
          m_assocAddr = ch.GetSrc();
          ScheduleData(pkt);
        }
      else if(m_state == RTSSENT)
          ScheduleData(pkt);
      else
        NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " received CTS while state != RTSSENT or GWPING");
      break;
    case TYPE_GWPING:
      //Do not respond to GWPINGS at non-gateway nodes
      break;
    case TYPE_ACK:
      m_rtsBlocked = true;
      if(ch.GetDest() != m_address)
        return;
      //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " UanMacRc Receiving ACK packet from PHY");
      ProcessAck(pkt);
      break;
    default:
      NS_FATAL_ERROR("Unknown packet type " << ch.GetType() << " received at node " << GetAddress());
  }

}

void
UanMacRc::ScheduleData(Ptr<Packet> cts)
{
  NS_ASSERT(m_state == RTSSENT || m_state == GWPSENT);

  UanHeaderRcCts ctsh;
  cts->RemoveHeader(ctsh);

  std::list<Reservation>::iterator it = m_resList.begin();
  for(;it != m_resList.end();it++)
    {
      if(it->GetFrameNo() == ctsh.GetFrameNo())
        break;
    }
  if(it == m_resList.end())
    {
      NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " received CTS packet with no corresponding reservation!");
      //NS_FATAL_ERROR("In " << __func__ << " could not find reservation corresponding to received CTS");
      return;
    }
  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " received CTS packet.  Scheduling data");
  it->SetTransmitted();

  double currentBps = m_phy->GetMode(m_currentRate).GetDataRateBps();

  //Since we're synchronized, just compute prop. delay with timestamp
  //We could synchronize by using the 4 timestamps if an offset existed
  Time propDelay = Simulator::Now() - ctsh.GetTimeStamp2() - Seconds(m_ctsSize*8 / currentBps);
  //NS_LOG_DEBUG("Calculated prop delay " << propDelay.GetSeconds());

  Time arrTime = ctsh.GetTimeStamp2() + ctsh.GetDelayToTx();
  Time txTime = arrTime - propDelay;

  //Time startDelay = ctsh.GetTimeStamp2() + ctsh.GetDelayToTx() - Simulator::Now() + ctsh.GetTimeStamp2()
  Time startDelay = txTime - Simulator::Now();

  Time frameDelay = Seconds(0);

  const std::list<std::pair <Ptr<Packet>, UanAddress > > l = it->GetPktList();
  std::list<std::pair <Ptr<Packet>, UanAddress > >::const_iterator pit;
  pit = l.begin();



  for(uint32_t i=0;i<it->GetNoFrames();i++, pit++)
    {
      Ptr<Packet> pkt = (*pit).first->Copy();
      //UanAddress dest = (*pit).second;
      UanHeaderRcData dh;
      dh.SetFrameNo(i);
      dh.SetPropDelay(propDelay);
      pkt->AddHeader(dh);
      UanHeaderCommon ch;
      ch.SetType(TYPE_DATA);
      //NS_ASSERT(m_assocAddr == dest);
      ch.SetDest(m_assocAddr);
      ch.SetSrc(m_address);
      pkt->AddHeader(ch);
      Time eventTime = startDelay + frameDelay;
      if(eventTime.GetSeconds() < 0)
        {
          if(eventTime.GetSeconds() > -0.002)
            eventTime = Seconds(0);
          else
            NS_FATAL_ERROR("Scheduling error resulted in very negative data transmission time! eventTime = " << eventTime.GetSeconds());
        }
      Simulator::Schedule(eventTime, &UanMacRc::SendPacket, this, pkt, m_currentRate);
      frameDelay = frameDelay + m_sifs + Seconds(pkt->GetSize()/currentBps);
    }

  m_state = IDLE;
  if(!m_pktQueue.empty())
    {

      if(m_rtsEvent.IsRunning())
        m_rtsEvent.Cancel();
	  ExponentialVariable ev(1/m_retryRate);
      double timeout = ev.GetValue();
      m_rtsEvent = Simulator::Schedule(Seconds(timeout), &UanMacRc::SendRts, this);
    }

}

void
UanMacRc::SendPacket(Ptr<Packet> pkt, uint32_t rate)
{
  UanHeaderCommon ch;
  pkt->PeekHeader(ch);
  std::string type;
  switch(ch.GetType())
  {
    case TYPE_DATA:
      type = "DATA";
      break;
    case TYPE_RTS:
      type= "RTS";
      break;
    case TYPE_CTS:
      type="CTS";
      break;
    case TYPE_ACK:
      type="ACK";
      break;
    case TYPE_GWPING:
      type="GWPING";
      break;
    default:
      type="UNKNOWN";
      break;
  }
  NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " transmitting " << pkt->GetSize() << " byte packet of type " << type << " with rate " << rate << "(" << m_phy->GetMode(rate).GetDataRateBps() << ") to " << ch.GetDest());
  m_dequeueLogger(pkt, rate);
  m_phy->SendPacket(pkt, rate);
}

void
UanMacRc::ProcessAck(Ptr<Packet> ack)
{
  UanHeaderRcAck ah;
  ack->RemoveHeader(ah);

  std::list<Reservation>::iterator it = m_resList.begin();
  for(;it != m_resList.end();it++)
    {
      if(it->GetFrameNo() == ah.GetFrameNo())
        break;
    }
  if(it == m_resList.end())
    {
      NS_LOG_DEBUG("In " << __func__ << " could not find reservation corresponding to received ACK");
      return;
    }
  if(!it->IsTransmitted())
    return;
  if(ah.GetNoNacks() > 0)
    {
      const std::list<std::pair <Ptr<Packet>, UanAddress > > l = it->GetPktList();
      std::list<std::pair <Ptr<Packet>, UanAddress > >::const_iterator pit;
      pit = l.begin();

      const std::set<uint8_t> &nacks = ah.GetNackedFrames();
      std::set<uint8_t>::iterator nit = nacks.begin();
      uint8_t pnum=0;
      for(;nit != nacks.end();nit++)
        {
          NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " Received NACK for " << (uint32_t) *nit);
          while(pnum < *nit)
            {
              pit++;
              pnum++;
            }
          UanHeaderRcData dh;
          UanHeaderCommon ch;
          m_pktQueue.push_front(*pit);
        }
    }
  else
    {
      NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Node " << m_address << " received ACK for all frames");
    }
  m_resList.erase(it);
//  if(m_resList.empty())
//    m_state = IDLE;
//  else
//    m_state = RTSSENT;

}

UanHeaderRcRts
UanMacRc::CreateRtsHeader(const Reservation &res)
{
  UanHeaderRcRts rh = UanHeaderRcRts();
  //NS_LOG_DEBUG("ResLength=" << res.GetLength() << " # Frames=" << res.GetNoFrames() << " Timestamp=" << res.GetTimestamp(res.GetRetryNo()) << " Frame #=" << res.GetFrameNo() << " Retry #=" << res.GetRetryNo());
  rh.SetLength(res.GetLength());
  rh.SetNoFrames(res.GetNoFrames());
  rh.SetTimeStamp(res.GetTimestamp(res.GetRetryNo()));
  rh.SetFrameNo(res.GetFrameNo());
  rh.SetRetryNo(res.GetRetryNo());

  //NS_LOG_DEBUG("Created RTS header " << rh);
  return rh;
}

void
UanMacRc::Associate(void)
{
  m_cntrlSends++;

  Reservation res(m_pktQueue, m_frameNo, m_maxFrames);
  res.AddTimestamp(Simulator::Now());
  m_frameNo++;
  m_resList.push_back(res);
  Ptr<UanPhyDual> phyDual = m_phy->GetObject<UanPhyDual>();
  bool phy1ok = IsPhy1Ok();
  if(phy1ok && !phyDual->IsPhy2Tx() & !m_rtsBlocked)
    {
      Ptr<Packet> pkt = Create<Packet>(0);
      pkt->AddHeader(CreateRtsHeader(res));
      pkt->AddHeader(UanHeaderCommon(m_address, UanAddress::GetBroadcast(), (uint8_t) TYPE_GWPING));
      NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Sending first GWPING " << *pkt);
      SendPacket(pkt,m_currentRate + m_numRates);
    }
  m_state = GWPSENT;
  NS_ASSERT(!m_rtsEvent.IsRunning());
  ExponentialVariable ev(1.0/m_retryRate);
  double timeout = ev.GetValue();
  m_rtsEvent = Simulator::Schedule(Seconds(timeout), &UanMacRc::AssociateTimeout, this);
}

void
UanMacRc::AssociateTimeout()
{
  m_cntrlSends++;
  if(m_state != GWPSENT)
    return;
  Ptr<UanPhyDual> phyDual = m_phy->GetObject<UanPhyDual>();
  bool phy1ok = IsPhy1Ok();
  if(phy1ok && !phyDual->IsPhy2Tx() && !m_rtsBlocked)
    {
      Ptr<Packet> pkt = Create<Packet>();


      Reservation res = m_resList.back();
      m_resList.pop_back();
      res.AddTimestamp(Simulator::Now());
      res.IncrementRetry();

      pkt->AddHeader(CreateRtsHeader(res));
      pkt->AddHeader(UanHeaderCommon(m_address, UanAddress::GetBroadcast(), (uint8_t) TYPE_GWPING));
      //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Sending retry GWPING " << *pkt);
      SendPacket(pkt,m_currentRate + m_numRates);
      m_resList.push_back(res);
    }
  NS_ASSERT(!m_rtsEvent.IsRunning());
  ExponentialVariable ev(1.0/m_retryRate);
  double timeout = ev.GetValue();
  m_rtsEvent = Simulator::Schedule(Seconds(timeout), &UanMacRc::AssociateTimeout, this);

}


void
UanMacRc::SendRts(void)
{
  m_cntrlSends++;
  if(m_state == RTSSENT)
    return;

  NS_ASSERT(!m_pktQueue.empty());

  Reservation res(m_pktQueue, m_frameNo, m_maxFrames);
  res.AddTimestamp(Simulator::Now());
  m_frameNo++;
  m_resList.push_back(res);
  Ptr<UanPhyDual> phyDual = m_phy->GetObject<UanPhyDual>();
  bool phy1ok = IsPhy1Ok();
  if(phy1ok && !phyDual->IsPhy2Tx() && !m_rtsBlocked )
    {
      Ptr<Packet> pkt = Create<Packet>(0);
      pkt->AddHeader(CreateRtsHeader(res));
      pkt->AddHeader(UanHeaderCommon(m_address, UanAddress::GetBroadcast(), (uint8_t) TYPE_RTS));
      SendPacket(pkt,m_currentRate + m_numRates);
    }
  m_state = RTSSENT;
  NS_ASSERT(!m_rtsEvent.IsRunning());
  ExponentialVariable ev(1.0/m_retryRate);
  double timeout = ev.GetValue();
  m_rtsEvent = Simulator::Schedule(Seconds(timeout), &UanMacRc::RtsTimeout, this);

}

bool
UanMacRc::IsPhy1Ok()
{
  Ptr<UanPhyDual> phyDual = m_phy->GetObject<UanPhyDual>();

  bool phy1ok = true;
  if(phyDual->IsPhy1Rx())
    {
      Ptr<Packet> pkt = phyDual->GetPhy1PacketRx();
      UanHeaderCommon ch;
      pkt->PeekHeader(ch);
      if(ch.GetType() == TYPE_CTS || ch.GetType() == TYPE_ACK)
        phy1ok = false;

    }
  return phy1ok;
}
void
UanMacRc::RtsTimeout(void)
{
  m_cntrlSends++;

  if(m_state != RTSSENT)
    return;
  Ptr<UanPhyDual> phyDual = m_phy->GetObject<UanPhyDual>();

  bool phy1ok = IsPhy1Ok();
  if(phy1ok && !phyDual->IsPhy2Tx() && !m_rtsBlocked)
    {

      if(m_resList.empty())
        {
          NS_FATAL_ERROR(Simulator::Now().GetSeconds() << " Node " << m_address << " tried to retry RTS with empty reservation list");
        }
      Ptr<Packet> pkt = Create<Packet>(0);

      Reservation res = m_resList.back();
      NS_ASSERT(!res.IsTransmitted());
      m_resList.pop_back();
      res.AddTimestamp(Simulator::Now());
      res.IncrementRetry();
      m_resList.push_back(res);
      pkt->AddHeader(CreateRtsHeader(res));
      pkt->AddHeader(UanHeaderCommon(m_address, UanAddress::GetBroadcast(), (uint8_t) TYPE_RTS));
      SendPacket(pkt,m_currentRate + m_numRates);

    }
  m_state = RTSSENT;
  NS_ASSERT(!m_rtsEvent.IsRunning());
  ExponentialVariable ev(1.0/m_retryRate);
  double timeout = ev.GetValue();
  m_rtsEvent = Simulator::Schedule(Seconds(timeout), &UanMacRc::RtsTimeout, this);
}

Reservation::Reservation()
  : m_length(0),
    m_frameNo(0),
    m_retryNo(0),
    m_transmitted(false)
{

}



Reservation::Reservation(std::list<std::pair <Ptr<Packet>, UanAddress > > &list, uint8_t frameNo, uint32_t maxPkts)
  : m_frameNo(frameNo),
    m_retryNo(0),
    m_transmitted(false)
{
  uint32_t numPkts = (maxPkts)?maxPkts:list.size();
  uint32_t length = 0;

  for(uint32_t i=0;i<numPkts;i++)
    {
      //+3 for size of data header +3 for size of common header
      length += list.front().first->GetSize() + 6;
      m_pktList.push_back(list.front());
      list.pop_front();
    }
  m_length = length;
}

}
