/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

#ifndef UANMACRC_H_
#define UANMACRC_H_

#include "uan-mac.h"
#include "uan-address.h"

#include "ns3/nstime.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/traced-callback.h"
#include "ns3/event-id.h"

#include <list>
#include <utility>
#include <vector>


namespace ns3
{

class Address;
class UanTxMode;
class UanHeaderRcRts;
class UanPhy;


/**
 * \class Reservation
 * \brief Stores reservation info for use in scheduling data channel by reservation channel MAC
 */
class Reservation
{
public:
  Reservation();
  Reservation(std::list<std::pair <Ptr<Packet>, UanAddress > > &list, uint8_t frameNo, uint32_t maxPkts=0);

  uint32_t GetNoFrames() const { return m_pktList.size(); }
  uint32_t GetLength() const { return m_length; }
  const std::list<std::pair <Ptr<Packet>, UanAddress > > &GetPktList(void) const { return m_pktList; }
  uint8_t GetFrameNo() const { return m_frameNo; }
  uint8_t GetRetryNo() const { return m_retryNo; };
  Time GetTimestamp(uint8_t n) const { return m_timestamp[n]; }
  bool IsTransmitted() const { return m_transmitted; }

  void SetFrameNo(uint8_t fn) { m_frameNo = fn; }
  void AddTimestamp(Time t) { m_timestamp.push_back(t); }
  void IncrementRetry() { m_retryNo++; }
  void SetTransmitted(bool t = true) { m_transmitted = true; }
private:
  std::list<std::pair <Ptr<Packet>, UanAddress > > m_pktList;
  uint32_t m_length;
  uint8_t m_frameNo;
  std::vector<Time> m_timestamp;
  uint8_t m_retryNo;
  bool m_transmitted;
};


/**
 * \class UanMacRc
 * \brief Non-gateway node MAC for reservation channel MAC protocol
 *
 * This class assumes a network topology where all traffic
 * is destined for a set of GW nodes which are connected via
 * some out of band (RF?) means.
 *
 * This is a work in progress.  Currently there is no support
 * for multihop networks and a single GW is assumed (reachable by all nodes).
 */
class UanMacRc : public ns3::UanMac
{
public:
  enum {TYPE_DATA, TYPE_GWPING, TYPE_RTS, TYPE_CTS, TYPE_ACK};
  UanMacRc();
  virtual ~UanMacRc();

  static TypeId GetTypeId(void);

  //Inherited virtual functions
  virtual Address GetAddress(void);
  virtual void SetAddress(UanAddress addr);
  virtual bool Enqueue(Ptr<Packet> packet, const Address &dest, uint16_t protocolNumber);
  virtual void SetForwardUpCb(Callback<void, Ptr<Packet>, const UanAddress&> cb);
  virtual void AttachPhy(Ptr<UanPhy> phy);
  virtual Address GetBroadcast(void) const;

private:



  enum State {UNASSOCIATED, GWPSENT, IDLE, RTSSENT, DATATX};

  State m_state;
  bool m_rtsBlocked;

  EventId m_startAgain;
  UanAddress m_address;
  double m_retryRate;
  UanAddress m_assocAddr;
  Ptr<UanPhy> m_phy;
  uint32_t m_numRates;
  uint32_t m_currentRate;
  uint32_t m_maxFrames;
  uint32_t m_queueLimit;
  uint8_t m_frameNo;
  Time m_sifs;

  uint16_t m_numRetryRates;
  double m_minRetryRate;
  double m_retryStep;

  uint32_t m_ctsSize;

  std::list<std::pair <Ptr<Packet>, UanAddress > > m_pktQueue;
  std::list<Reservation> m_resList;

  Callback<void, Ptr<Packet>, const UanAddress& > m_forwardUpCb;

  TracedCallback<Ptr<const Packet>, UanTxMode > m_rxLogger;
  TracedCallback<Ptr<const Packet>, uint16_t > m_enqueueLogger;
  TracedCallback<Ptr<const Packet>, uint16_t > m_dequeueLogger;

  EventId m_rtsEvent;
  void ReceiveOkFromPhy(Ptr<Packet>, double sinr, UanTxMode mode);

  void Associate(void);
  void AssociateTimeout(void);
  void SendRts(void);
  void RtsTimeout(void);
  UanHeaderRcRts CreateRtsHeader(const Reservation &res);
  void ScheduleData(Ptr<Packet> cts);
  void ProcessAck(Ptr<Packet> ack);
  void SendPacket(Ptr<Packet> pkt, uint32_t rate);
  bool IsPhy1Ok(void);
  void StartRtsing(void) { m_rtsBlocked = false; }
  void BlockRtsing(void) { m_rtsBlocked = true; }

};

}

#endif /* UANMACRC_H_ */
