/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

#ifndef UANPHYDUAL_H_
#define UANPHYDUAL_H_

#include "ns3/uan-phy.h"



namespace ns3
{

class UanTxMode;
class UanModesList;


/**
 * \class UanPhyCalcSinrDual
 * \brief Default SINR model for UanPhyDual
 *
 * Considers interfering packet power as additional ambient noise only
 * if there is overlap in frequency band as found from supplied UanTxMode.
 * If there is no overlap, then the packets are considered not to interfere.
 */
class UanPhyCalcSinrDual : public UanPhyCalcSinr
{
public:
  UanPhyCalcSinrDual();
  virtual ~UanPhyCalcSinrDual();
  static TypeId GetTypeId(void);
  virtual double CalcSinrDb(Ptr<Packet> pkt,
                         Time arrTime,
                         double rxPowerDb,
                         double ambNoiseDb,
                         UanTxMode mode,
                         UanPdp pdp,
                         const UanTransducer::ArrivalList &arrivalList
                         ) const;
};

/**
 * \class UanPhyDual
 *
 * A class that wraps two generic UAN Phy layers into a single PHY.
 * This is used to simulate two receivers (and transmitters) that use
 * the same front end hardware.  When attached to a UanTransducerHd,
 * this results in a net device able to transmit on one or two channels
 * simultaneously or receive on one or two channels simultaneously but
 * that cannot transmit and receive simultaneously.
 *
 * Many of the standard PHY functions here become ambiguous.  In most cases
 * information for "Phy1" are returned.
 */
class UanPhyDual : public UanPhy
{
public:
  UanPhyDual();
  virtual ~UanPhyDual();

  static TypeId GetTypeId();

  virtual void SendPacket(Ptr<Packet> pkt, uint32_t modeNum);

  /**
   * \brief Register a class to receive phy state change notifications
   * \param listener Class derived from UanPhyListener to receive notifications
   *
   * Note that, from UanPhyDual, you may receive duplicate
   * messages as underneath there are two generic phys here.
   * Each will notify of state changes independently.
   */
  virtual void RegisterListener(UanPhyListener *listener);
  virtual void StartRxPacket(Ptr<Packet> pkt, double rxPowerDb, UanTxMode txMode, UanPdp pdp);
  virtual void SetReceiveOkCallback (RxOkCallback cb);
  virtual void SetReceiveErrorCallback (RxErrCallback cb);
  virtual void SetRxGainDb(double gain);
  virtual void SetTxPowerDb(double txpwr);
  virtual void SetRxThresholdDb(double thresh);
  virtual void SetCcaThresholdDb(double thresh);
  virtual double GetRxGainDb(void);
  virtual double GetTxPowerDb(void);
  virtual double GetRxThresholdDb(void);
  virtual double GetCcaThresholdDb(void);
  virtual bool IsStateIdle(void);
  virtual bool IsStateBusy(void);
  virtual bool IsStateRx(void);
  virtual bool IsStateTx(void);
  virtual bool IsStateCcaBusy(void);
  virtual Ptr<UanChannel> GetChannel(void) const;
  virtual Ptr<UanNetDevice> GetDevice(void);
  virtual void SetChannel(Ptr<UanChannel> channel);
  virtual void SetDevice(Ptr<UanNetDevice> device);
  virtual void SetMac(Ptr<UanMac> mac);
  virtual void NotifyTransStartTx(Ptr<Packet> packet, double txPowerDb, UanTxMode txMode);
  virtual void NotifyIntChange(void);
  virtual void SetTransducer(Ptr<UanTransducer> trans);
  virtual Ptr<UanTransducer> GetTransducer(void);
  virtual uint32_t GetNModes(void);
  virtual UanTxMode GetMode(uint32_t n);

  bool IsPhy1Idle(void);
  bool IsPhy2Idle(void);
  bool IsPhy1Rx(void);
  bool IsPhy2Rx(void);
  bool IsPhy1Tx(void);
  bool IsPhy2Tx(void);

  //Attribute getters and setters
  double GetCcaThresholdPhy1(void) const;
  double GetCcaThresholdPhy2(void) const;
  void SetCcaThresholdPhy1(double);
  void SetCcaThresholdPhy2(double);

  double GetTxPowerDbPhy1(void) const;
  double GetTxPowerDbPhy2(void) const;
  void SetTxPowerDbPhy1(double);
  void SetTxPowerDbPhy2(double);

  double GetRxGainDbPhy1(void) const;
  double GetRxGainDbPhy2(void) const;
  void SetRxGainDbPhy1(double);
  void SetRxGainDbPhy2(double);

  /**
   * \returns List of available modes on Phy1
   */
  UanModesList GetModesPhy1(void) const;
  /**
   * \returns List of available modes on Phy2
   */
  UanModesList GetModesPhy2(void) const;
  /**
   * \param modes List of modes to use on Phy1 (index corresponds to mode #)
   */
  void SetModesPhy1(UanModesList modes);
  /*
   * \param modes List of modes to use on Phy2 (index corresponds to mode #)
   */
  void SetModesPhy2(UanModesList modes);

  /**
   * \returns Ptr to PER model for Phy1
   */
  Ptr<UanPhyPer> GetPerFunctorPhy1(void) const;
  /**
   * \returns Ptr to PER model for Phy2
   */
  Ptr<UanPhyPer> GetPerFunctorPhy2(void) const;
  /**
   * \param per Ptr to PER model to use on Phy1
   */
  void SetPerFunctorPhy1(Ptr<UanPhyPer> per);
  /**
   * \param per Ptr to PER model to use on Phy2
   */
  void SetPerFunctorPhy2(Ptr<UanPhyPer> per);

  /**
   * \returns Ptr to SINR model for Phy1
   */
  Ptr<UanPhyCalcSinr> GetSinrFunctorPhy1(void) const;
  /*
   * \returns Ptr to SINR model for Phy2
   */
  Ptr<UanPhyCalcSinr> GetSinrFunctorPhy2(void) const;
  /**
   * \param calcSinr Ptr to SINR model to use on Phy1
   */
  void SetSinrFunctorPhy1(Ptr<UanPhyCalcSinr> calcSinr);
  /**
   * \param calcSinr Ptr to SINR model to use on Phy2
   */
  void SetSinrFunctorPhy2(Ptr<UanPhyCalcSinr> calcSinr);

  /**
   * \returns Packet currently being received on Phy1 (Null Ptr if none)
   */
  Ptr<Packet> GetPhy1PacketRx(void) const;
  /**
   * \returns Packet currently being received on Phy2 (Null Ptr if none)
   */
  Ptr<Packet> GetPhy2PacketRx(void) const;
  /**
   * \returns Packet currenty being received on Phy1 (Null Ptr if none)
   */
  Ptr<Packet> GetPacketRx(void) const;

private:

  Ptr<UanPhy> m_phy1;
  Ptr<UanPhy> m_phy2;
  TracedCallback<Ptr<const Packet>, double, UanTxMode > m_rxOkLogger;
  TracedCallback<Ptr<const Packet>, double, UanTxMode > m_rxErrLogger;
  TracedCallback<Ptr<const Packet>, double, UanTxMode > m_txLogger;
  RxOkCallback m_recOkCb;
  RxErrCallback m_recErrCb;


  void RxOkFromSubPhy(Ptr<Packet> pkt, double sinr, UanTxMode mode);
  void RxErrFromSubPhy(Ptr<Packet> pkt, double sinr);

};

}

#endif /* UANPHYDUAL_H_ */
