/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

#include "uan-phy-gen.h"
#include "uan-transducer.h"
#include "uan-channel.h"
#include "uan-net-device.h"
#include "ns3/simulator.h"
#include "ns3/traced-callback.h"
#include "ns3/ptr.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/double.h"
#include "ns3/log.h"
#include "ns3/uan-tx-mode.h"
#include "ns3/node.h"
#include "ns3/uinteger.h"
#include "ns3/random-variable.h"

#include <iostream>
NS_LOG_COMPONENT_DEFINE("UanPhyGen");

namespace ns3
{

NS_OBJECT_ENSURE_REGISTERED (UanPhyGen);
NS_OBJECT_ENSURE_REGISTERED (UanPhyPerGenDefault);
NS_OBJECT_ENSURE_REGISTERED (UanPhyCalcSinrDefault);
NS_OBJECT_ENSURE_REGISTERED (UanPhyCalcSinrFhFsk);
NS_OBJECT_ENSURE_REGISTERED (UanPhyPerUmodem);

UanPhyCalcSinrDefault::UanPhyCalcSinrDefault()
{

}
UanPhyCalcSinrDefault::~UanPhyCalcSinrDefault()
{

}

TypeId
UanPhyCalcSinrDefault::GetTypeId(void)
{
  static TypeId tid = TypeId("ns3::UanPhyCalcSinrDefault")
    .SetParent<Object>()
    .AddConstructor<UanPhyCalcSinrDefault>()
  ;
  return tid;
}

double
UanPhyCalcSinrDefault::CalcSinrDb(Ptr<Packet> pkt,
                         Time arrTime,
                         double rxPowerDb,
                         double ambNoiseDb,
                         UanTxMode mode,
                         UanPdp pdp,
                         const UanTransducer::ArrivalList &arrivalList) const
{
  if(mode.GetModType() == UanTxMode::OTHER)
    {
      NS_LOG_WARN("Calculating SINR for unsupported modulation type");
    }

  //Ideally here, you could implement modulation specific models to
  //calculate SINR of incoming packets.
  //e.g.  If two packets are separated in frequency, or
  //are using orthogonal hopping codes in FH-FSK or whatever, or in some other
  //way occupy (nearly) orthogonal channels then the interference calculation
  //should reflect this.

  //The default ignores mode data and assumes that all rxpower transmitted is
  //captured by the receiver, and that all signal power associated with
  //interfering packets effects SINR identically to additional ambient noise.

  double intKp = -DbToKp(rxPowerDb); //This packet is in the arrivalList
  UanTransducer::ArrivalList::const_iterator it = arrivalList.begin();
  for(;it != arrivalList.end();it++)
    {
      intKp += DbToKp(it->GetRxPowerDb());
    }

  double totalIntDb = KpToDb(intKp + DbToKp(ambNoiseDb));

  NS_LOG_DEBUG("Calculating SINR:  RxPower = " << rxPowerDb << " dB.  Number of interferers = " << arrivalList.size() << "  Interference + noise power = " << totalIntDb << " dB.  SINR = " << rxPowerDb - totalIntDb << " dB.");
  return rxPowerDb - totalIntDb;
}


UanPhyCalcSinrFhFsk::UanPhyCalcSinrFhFsk()
{

}
UanPhyCalcSinrFhFsk::~UanPhyCalcSinrFhFsk()
{

}

TypeId
UanPhyCalcSinrFhFsk::GetTypeId(void)
{
  static TypeId tid = TypeId("ns3::UanPhyCalcSinrFhFsk")
    .SetParent<Object>()
    .AddConstructor<UanPhyCalcSinrFhFsk>()
    .AddAttribute("NumberOfHops",
                  "Number of frequencies in hopping pattern",
                  UintegerValue(13),
                  MakeUintegerAccessor(&UanPhyCalcSinrFhFsk::m_hops),
                  MakeUintegerChecker<uint32_t>())
  ;
  return tid;
}
double
UanPhyCalcSinrFhFsk::CalcSinrDb(Ptr<Packet> pkt,
                         Time arrTime,
                         double rxPowerDb,
                         double ambNoiseDb,
                         UanTxMode mode,
                         UanPdp pdp,
                         const UanTransducer::ArrivalList &arrivalList) const
{
  if(mode.GetModType() != UanTxMode::FSK)
    NS_LOG_WARN("Calculating SINR for unsupported mode type");



  double ts = 1.0 / mode.GetPhyRateSps();
  double clearingTime = (m_hops-1.0)*ts;

  double csp = pdp.SumTapsNc(Seconds(0), Seconds(ts));
  NS_LOG_DEBUG("Symbol duration = " << ts << " Captured signal power: " << csp);
  UanPdp::Iterator tmpIt = pdp.GetBegin();
//  for(;tmpIt != pdp.GetEnd();tmpIt++)
//    {
//      NS_LOG_DEBUG("Delay = " << tmpIt->GetDelay().GetSeconds() << " amp = " << tmpIt->GetAmp());
//    }

  double effRxPowerDb = rxPowerDb + KpToDb(csp);

  double isiKpa = DbToKp(rxPowerDb)*pdp.SumTapsNc(Seconds(ts+clearingTime), Seconds(2.0*ts+clearingTime));

  UanTransducer::ArrivalList::const_iterator it = arrivalList.begin();
  double intKp = -DbToKp(effRxPowerDb);
  for(;it != arrivalList.end();it++)
    {
      UanPdp intPdp = it->GetPdp();
      double tDelta = std::abs(arrTime.GetSeconds() - it->GetArrivalTime().GetSeconds());
      //We want tDelta in terms of a single symbol (i.e. if tDelta = 7.3 symbol+clearing
      //times, the offset in terms of the arriving symbol power is
      //0.3 symbol+clearing times.

      int32_t syms = (uint32_t) ( (double) tDelta / (ts + clearingTime));
      tDelta = tDelta - syms*(ts + clearingTime);

      //Align to pktRx
      if(arrTime  < it->GetArrivalTime())
        {
          tDelta = ts+clearingTime - tDelta;
        }

      double intPower = 0.0;
      if(tDelta < ts)
        {
          intPower += intPdp.SumTapsNc(Seconds(0), Seconds(ts - tDelta));
          intPower += intPdp.SumTapsNc(Seconds(ts - tDelta + clearingTime),
                                       Seconds(2*ts - tDelta + clearingTime));
        }
      else
        {
          Time start = Seconds(ts + clearingTime - tDelta);
          Time end = start + Seconds(ts);
          intPower += intPdp.SumTapsNc(start, end);

          start = start + Seconds(ts+clearingTime);
          end = start + Seconds(ts);
          intPower += intPdp.SumTapsNc(start, end);
        }
      intKp += DbToKp(it->GetRxPowerDb()) * intPower;
    }

  double totalIntDb = KpToDb(isiKpa + intKp + DbToKp(ambNoiseDb));

  NS_LOG_DEBUG("Calculating SINR:  RxPower = " << rxPowerDb << " dB.  Effective Rx power " << effRxPowerDb << " dB.  Number of interferers = " << arrivalList.size() << "  Interference + noise power = " << totalIntDb << " dB.  SINR = " << effRxPowerDb - totalIntDb << " dB.");
  return effRxPowerDb - totalIntDb;
}




UanPhyPerGenDefault::UanPhyPerGenDefault()
{

}

UanPhyPerGenDefault::~UanPhyPerGenDefault()
{

}
TypeId
UanPhyPerGenDefault::GetTypeId(void)
{
  static TypeId tid = TypeId("ns3::UanPhyPerGenDefault")
    .SetParent<Object>()
    .AddConstructor<UanPhyPerGenDefault>()
    .AddAttribute("Threshold", "SINR cutoff for good packet reception",
                  DoubleValue(8),
                  MakeDoubleAccessor(&UanPhyPerGenDefault::m_thresh),
                  MakeDoubleChecker<double>());
  ;
  return tid;
}


//Default PER calculation simply compares SINR to a threshold which is configurable
//via an attribute.
double
UanPhyPerGenDefault::CalcPer(double sinrDb, UanTxMode mode)
{
  if(sinrDb>=m_thresh)
    {
      return 0;
    }
  else
    {
      return 1;
    }
}


UanPhyPerUmodem::UanPhyPerUmodem()
{

}
UanPhyPerUmodem::~UanPhyPerUmodem()
{

}

TypeId UanPhyPerUmodem::GetTypeId(void)
{
  static TypeId tid =
      TypeId("ns3::UanPhyPerUmodem")
          .SetParent<Object> ()
          .AddConstructor<UanPhyPerUmodem> ()
      ;
  return tid;
}

double
UanPhyPerUmodem::nchoosek(uint32_t n, uint32_t k)
{
  double result;

    result=1.0;

    for (uint32_t i = std::max(k,n-k) + 1; i <= n; ++i)
            result *= i;

    for (uint32_t i = 2; i <= std::min(k,n-k); ++i)
            result /= i;

    return result;
}

double
UanPhyPerUmodem::CalcPer(double sinr, UanTxMode mode)
{
  uint32_t d[] =
    { 12, 14, 16, 18, 20, 22, 24, 26, 28 };
  double Bd[] =
    { 33, 281, 2179, 15035LLU, 105166LLU, 692330LLU, 4580007LLU, 29692894LLU,
      190453145LLU };

  //double Rc = 1.0 / 2.0;
  double ebno = std::pow(10.0, sinr / 10.0);
  double perror = 1.0 / (2.0 + ebno);
  double P[9];

  if (sinr >= 10)
    return 0;
  if (sinr <= 6)
    return 1;

  for (uint32_t r = 0; r < 9; r++)
    {
      double sumd = 0;
      for (uint32_t k = 0; k < d[r]; k++)
        {
          sumd = sumd + nchoosek(d[r] - 1 + k, k) * std::pow(1 - perror, (double) k);
        }
      P[r] = std::pow(perror, (double) d[r]) * sumd;

    }

  double Pb = 0;
  for (uint32_t r = 0; r < 8; r++)
    {
      Pb = Pb + Bd[r] * P[r];
    }

  //cout << "Pb = " << Pb << endl;

  double Ppacket = 1;
  double temp = nchoosek(288, 0);
  temp *= std::pow((1 - Pb), 288.0);
  Ppacket -= temp;
  temp = nchoosek(288, 1) * Pb * std::pow((1 - Pb), 287);
  Ppacket -= temp;

  //NS_LOG_DEBUG("Pb = " << Pb << " Prob of Packet error = " << Ppacket);

  if (Ppacket > 1)
    return 1;
  else
    return Ppacket;
}

UanModesList
UanPhyGen::GetDefaultModes(void)
{
  UanModesList l;


  l.AppendMode(UanTxModeFactory::CreateMode(UanTxMode::FSK,80,80,22000,4000,13,"FSK"));
  l.AppendMode(UanTxModeFactory::CreateMode(UanTxMode::PSK,200, 200, 22000, 4000, 4, "QPSK"));
  return l;
}
UanPhyGen::UanPhyGen() :
  UanPhy(),
  m_state(IDLE),
  m_channel(0),
  m_transducer(0),
  m_device(0),
  m_mac(0),
  m_rxGainDb(0),
  m_txPwrDb(0),
  m_rxThreshDb(0),
  m_ccaThreshDb(0),
  m_pktRx(0)
{

}

UanPhyGen::~UanPhyGen()
{

}

TypeId
UanPhyGen::GetTypeId(void)
{

  static TypeId tid = TypeId("ns3::UanPhyGen")
    .SetParent<Object>()
    .AddConstructor<UanPhyGen>()
    .AddAttribute("CcaThreshold",
                  "Aggregate energy of incoming signals to move to CCA Busy state dB",
                  DoubleValue(10),
                  MakeDoubleAccessor(&UanPhyGen::m_ccaThreshDb),
                  MakeDoubleChecker<double>())
    .AddAttribute("RxThreshold",
                  "Required SNR for signal acquisition in dB",
                  DoubleValue(10),
                  MakeDoubleAccessor(&UanPhyGen::m_rxThreshDb),
                  MakeDoubleChecker<double>())
    .AddAttribute("TxPower",
                  "Transmission output power in dB",
                  DoubleValue(190),
                  MakeDoubleAccessor(&UanPhyGen::m_txPwrDb),
                  MakeDoubleChecker<double>())
    .AddAttribute("RxGain",
                  "Gain added to incoming signal at receiver",
                  DoubleValue(0),
                  MakeDoubleAccessor(&UanPhyGen::m_rxGainDb),
                  MakeDoubleChecker<double>())
    .AddAttribute("SupportedModes",
                  "List of modes supported by this PHY",
                  UanModesListValue(UanPhyGen::GetDefaultModes()),
                  MakeUanModesListAccessor(&UanPhyGen::m_modes),
                  MakeUanModesListChecker() )
    .AddAttribute("PerFunctor",
                  "Functor to calculate PER based on SINR and TxMode",
                  PointerValue(CreateObject<UanPhyPerGenDefault>()),
                  MakePointerAccessor(&UanPhyGen::m_per),
                  MakePointerChecker<UanPhyPer>())
    .AddAttribute("SinrFunctor",
                  "Functor to calculate SINR based on pkt arrivals and modes",
                  PointerValue(CreateObject<UanPhyCalcSinrDefault>()),
                  MakePointerAccessor(&UanPhyGen::m_sinr),
                  MakePointerChecker<UanPhyCalcSinr>())
    .AddTraceSource("RxOk",
                    "A packet was received successfully",
                    MakeTraceSourceAccessor(&UanPhyGen::m_rxOkLogger))
    .AddTraceSource("RxError",
                    "A packet was received unsuccessfully",
                    MakeTraceSourceAccessor(&UanPhyGen::m_rxErrLogger))
    .AddTraceSource("Tx",
                    "Packet transmission beginning",
                    MakeTraceSourceAccessor(&UanPhyGen::m_txLogger))
    ;
  return tid;

}
void
UanPhyGen::SendPacket(Ptr<Packet> pkt, uint32_t modeNum)
{
  NS_LOG_DEBUG("PHY " << m_mac->GetAddress() << ": Transmitting packet");
  if(m_state == TX)
    {
      NS_LOG_DEBUG("PHY requested to TX while already Transmitting.  Dropping packet.");
      return;
    }

  UanTxMode txMode = GetMode(modeNum);

  if(m_pktRx != 0)
    {
      m_minRxSinrDb = -1e30;
      m_pktRx = 0;
    }

  m_transducer->Transmit(Ptr<UanPhy>(this), pkt, m_txPwrDb, txMode);
  m_state = TX;
  double txdelay = pkt->GetSize()*8.0/txMode.GetDataRateBps();
  Simulator::Schedule(Seconds(txdelay), &UanPhyGen::TxEndEvent, this);
  NS_LOG_DEBUG("PHY " << m_mac->GetAddress() << " notifying listeners");
  NotifyListenersTxStart(Seconds(txdelay));
  m_txLogger(pkt, m_txPwrDb, txMode);
}

void
UanPhyGen::TxEndEvent()
{
  NS_ASSERT(m_state = TX);
  if(GetInterferenceDb((Ptr<Packet>) 0) > m_ccaThreshDb)
    {
      m_state = CCABUSY;
      NotifyListenersCcaStart();
    }
  else
    m_state = IDLE;
}

void
UanPhyGen::RegisterListener(UanPhyListener *listener)
{
  m_listeners.push_back(listener);
}


void
UanPhyGen::StartRxPacket(Ptr<Packet> pkt, double rxPowerDb, UanTxMode txMode, UanPdp pdp)
{

  //TODO:  Log new arriving packet
  switch(m_state)
  {
    case TX:
      NS_ASSERT(false);
      break;
    case RX:
      {
        NS_ASSERT(m_pktRx);
        double newSinrDb = CalculateSinrDb(m_pktRx, m_pktRxArrTime, m_rxRecvPwrDb, m_pktRxMode, m_pktRxPdp);
        m_minRxSinrDb  = (newSinrDb < m_minRxSinrDb)?newSinrDb:m_minRxSinrDb;
        NS_LOG_DEBUG("PHY " << m_mac->GetAddress() << ": Starting RX in RX mode.  SINR of pktRx = " << m_minRxSinrDb);
      }
      break;

    case CCABUSY:
    case IDLE:
      {
        NS_ASSERT(!m_pktRx);
        bool hasmode = false;
        for(uint32_t i=0;i<GetNModes();i++)
          {
            if(txMode.GetUid() == GetMode(i).GetUid())
              {
                hasmode = true;
                break;
              }
          }
        if(!hasmode)
          {
            //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Skipping unknown mod. packet");
            break;
          }


        double newsinr = CalculateSinrDb(pkt, Simulator::Now(), rxPowerDb, txMode, pdp);
        NS_LOG_DEBUG("PHY " << m_mac->GetAddress() << ": Starting RX in IDLE mode.  SINR = " << newsinr);
        if(newsinr > m_rxThreshDb)
          {
            m_state=RX;
            m_rxRecvPwrDb = rxPowerDb;
            m_minRxSinrDb = newsinr;
            m_pktRx = pkt;
            m_pktRxArrTime = Simulator::Now();
            m_pktRxMode = txMode;
            m_pktRxPdp = pdp;
            double txdelay = pkt->GetSize()*8.0/txMode.GetDataRateBps();
            Simulator::Schedule(Seconds(txdelay), &UanPhyGen::RxEndEvent, this, pkt, rxPowerDb, txMode);
            NotifyListenersRxStart();
          }

      }
      break;
  }

  if(m_state==IDLE && GetInterferenceDb((Ptr<Packet>) 0) > m_ccaThreshDb)
    {
      m_state = CCABUSY;
      NotifyListenersCcaStart();
    }

}

void
UanPhyGen::RxEndEvent(Ptr<Packet> pkt, double rxPowerDb, UanTxMode txMode)
{
  if(pkt != m_pktRx)
    return;

  if(GetInterferenceDb((Ptr<Packet>) 0) > m_ccaThreshDb)
    {
      m_state = CCABUSY;
      NotifyListenersCcaStart();
    }
  else
    m_state = IDLE;

  UniformVariable pg;

  if(pg.GetValue(0, 1) > m_per->CalcPer(m_minRxSinrDb, txMode))
    {
      m_rxOkLogger(pkt, m_minRxSinrDb, txMode);
      NotifyListenersRxGood();
      if(!m_recOkCb.IsNull())
        m_recOkCb(pkt, m_minRxSinrDb, txMode);

    }
  else
    {
      m_rxErrLogger(pkt, m_minRxSinrDb, txMode);
      NotifyListenersRxBad();
      if(!m_recErrCb.IsNull())
        m_recErrCb(pkt, m_minRxSinrDb);
    }

  m_pktRx=0;
}

void
UanPhyGen::SetReceiveOkCallback (RxOkCallback cb)
{
  m_recOkCb = cb;
}

void
UanPhyGen::SetReceiveErrorCallback (RxErrCallback cb)
{
  m_recErrCb = cb;
}


bool
UanPhyGen::IsStateIdle(void)
{
  return m_state==IDLE;
}
bool
UanPhyGen::IsStateBusy(void)
{
  return m_state != IDLE;
}
bool
UanPhyGen::IsStateRx(void)
{
  return m_state == RX;
}
bool
UanPhyGen::IsStateTx(void)
{
  return m_state == TX;
}

bool
UanPhyGen::IsStateCcaBusy(void)
{
  return m_state == CCABUSY;
}


void
UanPhyGen::SetRxGainDb(double gain)
{
  m_rxGainDb = gain;

}
void
UanPhyGen::SetTxPowerDb(double txpwr)
{
  m_txPwrDb = txpwr;
}
void
UanPhyGen::SetRxThresholdDb(double thresh)
{
  m_rxThreshDb = thresh;
}
void
UanPhyGen::SetCcaThresholdDb(double thresh)
{
  m_ccaThreshDb = thresh;
}
double
UanPhyGen::GetRxGainDb(void)
{
  return m_rxGainDb;
}
double
UanPhyGen::GetTxPowerDb(void)
{
  return m_txPwrDb;

}
double
UanPhyGen::GetRxThresholdDb(void)
{
  return m_rxThreshDb;
}
double
UanPhyGen::GetCcaThresholdDb(void)
{
  return m_ccaThreshDb;
}

Ptr<UanChannel>
UanPhyGen::GetChannel(void) const
{
  return m_channel;
}

Ptr<UanNetDevice>
UanPhyGen::GetDevice(void)
{
  return m_device;
}

Ptr<UanTransducer>
UanPhyGen::GetTransducer(void)
{
  return m_transducer;
}
void
UanPhyGen::SetChannel(Ptr<UanChannel> channel)
{
  m_channel = channel;
}

void
UanPhyGen::SetDevice(Ptr<UanNetDevice> device)
{
  m_device = device;
}

void
UanPhyGen::SetMac(Ptr<UanMac> mac)
{
  m_mac = mac;
}

void
UanPhyGen::SetTransducer(Ptr<UanTransducer> trans)
{
  m_transducer = trans;
  m_transducer->AddPhy(this);
}


void
UanPhyGen::NotifyTransStartTx(Ptr<Packet> packet, double txPowerDb, UanTxMode txMode)
{
  if(m_pktRx)
    m_minRxSinrDb = -1e30;
}

void
UanPhyGen::NotifyIntChange(void)
{
  if(m_state == CCABUSY && GetInterferenceDb((Ptr<Packet>) 0) < m_ccaThreshDb)
    {
      m_state = IDLE;
      NotifyListenersCcaEnd();
    }
}

double
UanPhyGen::CalculateSinrDb(Ptr<Packet> pkt, Time arrTime, double rxPowerDb, UanTxMode mode, UanPdp pdp)
{
  double noiseDb = m_channel->GetNoiseDbHz(mode.GetCenterFreqHz()) + 10*log10(mode.GetBandwidthHz());
  return m_sinr->CalcSinrDb(pkt, arrTime, rxPowerDb, noiseDb, mode, pdp, m_transducer->GetArrivalList());
}

double
UanPhyGen::GetInterferenceDb(Ptr<Packet> pkt)
{

  const UanTransducer::ArrivalList &arrivalList = m_transducer->GetArrivalList();

  UanTransducer::ArrivalList::const_iterator it = arrivalList.begin();

  double interfPower=0;

  for(;it != arrivalList.end();it++)
    {
      if(pkt != it->GetPacket())
        interfPower += DbToKp(it->GetRxPowerDb());
    }

  return KpToDb(interfPower);

}

double
UanPhyGen::DbToKp(double db)
{
  return pow(10, db/10.0);
}
double
UanPhyGen::KpToDb(double kp)
{
  return 10*log10(kp);
}

void
UanPhyGen::NotifyListenersRxStart(void)
{
  ListenerList::const_iterator it = m_listeners.begin();
  for(;it!=m_listeners.end();it++)
    {
      (*it)->NotifyRxStart();
    }

}
void
UanPhyGen::NotifyListenersRxGood(void)
{
  ListenerList::const_iterator it = m_listeners.begin();
  for(;it!=m_listeners.end();it++)
    {
      (*it)->NotifyRxEndOk();
    }
}
void
UanPhyGen::NotifyListenersRxBad(void)
{
  ListenerList::const_iterator it = m_listeners.begin();
  for(;it!=m_listeners.end();it++)
    {
      (*it)->NotifyRxEndError();
    }
}
void
UanPhyGen::NotifyListenersCcaStart(void)
{
  ListenerList::const_iterator it = m_listeners.begin();
  for(;it!=m_listeners.end();it++)
    {
      (*it)->NotifyCcaStart();
    }
}
void
UanPhyGen::NotifyListenersCcaEnd(void)
{
  ListenerList::const_iterator it = m_listeners.begin();
  for(;it!=m_listeners.end();it++)
    {
      (*it)->NotifyCcaEnd();
    }
}

void
UanPhyGen::NotifyListenersTxStart(Time duration)
{
  ListenerList::const_iterator it = m_listeners.begin();
  for(;it!=m_listeners.end();it++)
    {
      (*it)->NotifyTxStart(duration);
    }
}

uint32_t
UanPhyGen::GetNModes(void)
{
  return m_modes.GetNModes();
}

UanTxMode
UanPhyGen::GetMode(uint32_t n)
{
  NS_ASSERT(n < m_modes.GetNModes());

  return m_modes[n];
}

Ptr<Packet>
UanPhyGen::GetPacketRx(void) const
{
  return m_pktRx;
}


} // namespace ns3
