
#include "uan-phy-simple.h"
#include "uan-transducer.h"
#include "uan-channel.h"
#include "uan-net-device.h"
#include "ns3/simulator.h"
#include "ns3/traced-callback.h"
#include "ns3/ptr.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/double.h"
#include "ns3/log.h"

#include <iostream>
NS_LOG_COMPONENT_DEFINE("UanPhySimple");

namespace ns3
{

NS_OBJECT_ENSURE_REGISTERED (UanPhySimple);

UanPhySimple::UanPhySimple() :
  UanPhy(),
  m_state(IDLE),
  m_channel(0),
  m_transducer(0),
  m_device(0),
  m_mac(0),
  m_rxGainDb(0),
  m_txPwrDb(0),
  m_rxThreshDb(0),
  m_ccaThreshDb(0),
  m_pktRx(0)
{
  m_modes.AppendMode(UanTxModeFactory::CreateMode(UanTxMode::FSK,80,80,22000,4000,13,"FSK"));

}

UanPhySimple::~UanPhySimple()
{

}

TypeId
UanPhySimple::GetTypeId(void)
{

  static TypeId tid = TypeId("ns3::UanPhySimple")
    .SetParent<Object>()
    .AddConstructor<UanPhySimple>()
    .AddAttribute("CcaThreshold",
                  "Aggregate energy of incoming signals to move to CCA Busy state dB",
                  DoubleValue(10),
                  MakeDoubleAccessor(&UanPhySimple::m_ccaThreshDb),
                  MakeDoubleChecker<double>())
    .AddAttribute("RxThreshold",
                  "Required SNR for signal acquisition in dB",
                  DoubleValue(10),
                  MakeDoubleAccessor(&UanPhySimple::m_rxThreshDb),
                  MakeDoubleChecker<double>())
    .AddAttribute("TxPower",
                  "Transmission output power in dB",
                  DoubleValue(190),
                  MakeDoubleAccessor(&UanPhySimple::m_txPwrDb),
                  MakeDoubleChecker<double>())
    .AddAttribute("RxGain",
                  "Gain added to incoming signal at receiver",
                  DoubleValue(0),
                  MakeDoubleAccessor(&UanPhySimple::m_rxGainDb),
                  MakeDoubleChecker<double>())
    .AddTraceSource("RxOk",
                    "A packet was received successfully",
                    MakeTraceSourceAccessor(&UanPhySimple::m_rxOkLogger))
    .AddTraceSource("RxError",
                    "A packet was received unsuccessfully",
                    MakeTraceSourceAccessor(&UanPhySimple::m_rxErrLogger))
    .AddTraceSource("Tx",
                    "Packet transmission beginning",
                    MakeTraceSourceAccessor(&UanPhySimple::m_txLogger))
    ;
  return tid;

}
void
UanPhySimple::SendPacket(Ptr<Packet> pkt, uint32_t modeNum)
{
  //TODO:  Do something different here

  //NS_LOG_DEBUG("Phy Sending packet");
  if(m_state == TX)
    return;

  UanTxMode txMode = GetMode(modeNum);

  if(m_pktRx != 0)
    {
      m_minRxSinrDb = -1000;
    }

  m_transducer->Transmit(Ptr<UanPhy>(this), pkt, m_txPwrDb, txMode);
  m_state = TX;
  double txdelay = pkt->GetSize()*8.0/txMode.GetDataRateBps();
  Simulator::Schedule(Seconds(txdelay), &UanPhySimple::TxEndEvent, this);
  NotifyListenersTxStart(Seconds(txdelay));
  m_txLogger(pkt, m_txPwrDb, txMode);
}

void
UanPhySimple::TxEndEvent()
{
  if(GetInterferenceDb((Ptr<Packet>) 0) > m_ccaThreshDb)
    {
      m_state = CCABUSY;
      NotifyListenersCcaStart();
    }
  else
    m_state = IDLE;
}

void
UanPhySimple::RegisterListener(UanPhyListener *listener)
{
  m_listeners.push_back(listener);
}


void
UanPhySimple::StartRxPacket(Ptr<Packet> pkt, double rxPowerDb, UanTxMode txMode, UanPdp pdp)
{

  NS_LOG_DEBUG("Phy Starting RX");

  //TODO:  Log new arriving packet
  switch(m_state)
  {
    case TX:
      NS_ASSERT(false);
      break;
    case RX:
      {
        NS_ASSERT(m_pktRx);
        double newSinrDb = CalculateSinrDb(m_pktRx, m_rxRecvPwrDb);
        m_minRxSinrDb  = (newSinrDb < m_minRxSinrDb)?newSinrDb:m_minRxSinrDb;
      }
      break;

    case CCABUSY:
    case IDLE:
      {
        NS_ASSERT(!m_pktRx);
        double newsinr = CalculateSinrDb(m_pktRx, rxPowerDb);
        if(newsinr > m_rxThreshDb)
          {
            m_state=RX;
            m_rxRecvPwrDb = rxPowerDb;
            m_minRxSinrDb = newsinr;
            m_pktRx = pkt;
            double txdelay = pkt->GetSize()*8.0/txMode.GetDataRateBps();
            Simulator::Schedule(Seconds(txdelay), &UanPhySimple::RxEndEvent, this, pkt, rxPowerDb, txMode);
            NotifyListenersRxStart();
          }
      }
      break;
  }

  if(m_state==IDLE && GetInterferenceDb((Ptr<Packet>) 0) > m_ccaThreshDb)
    {
      m_state = CCABUSY;
      NotifyListenersCcaStart();
    }

}

void
UanPhySimple::RxEndEvent(Ptr<Packet> pkt, double rxPowerDb, UanTxMode txMode)
{
  if(pkt != m_pktRx)
    return;
  /*
   * Copyright (c) 2008 University of Washington
   *
   * This program is free software: you can redistribute it and/or modify
   * it under the terms of the GNU General Public License as published by
   * the Free Software Foundation, either version 3 of the License, or
   * (at your option) any later version.
   *
   * This program is distributed in the hope that it will be useful,
   * but WITHOUT ANY WARRANTY; without even the implied warranty of
   * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   * GNU General Public License for more details.
   *
   * You should have received a copy of the GNU General Public License
   * along with this program.  If not, see <http://www.gnu.org/licenses/>.
   *
   * Author: Leonard Tracy <lentracy@u.washington.edu>
   *
   *
   */


  if(m_minRxSinrDb > m_rxThreshDb)
    {
      m_rxOkLogger(pkt, m_minRxSinrDb, txMode);
      NotifyListenersRxGood();
      m_recOkCb(pkt, m_minRxSinrDb, txMode);

    }
  else
    {
	  m_rxOkLogger(pkt, m_minRxSinrDb, txMode);
      NotifyListenersRxBad();
      m_recErrCb(pkt, m_minRxSinrDb);


    }
  m_state = IDLE;
  m_pktRx=0;
}

void
UanPhySimple::SetReceiveOkCallback (RxOkCallback cb)
{
  m_recOkCb = cb;
}

void
UanPhySimple::SetReceiveErrorCallback (RxErrCallback cb)
{
  m_recErrCb = cb;
}


bool
UanPhySimple::IsStateIdle(void)
{
  return m_state==IDLE;
}
bool
UanPhySimple::IsStateBusy(void)
{
  return m_state != IDLE;
}
bool
UanPhySimple::IsStateRx(void)
{
  return m_state == RX;
}
bool
UanPhySimple::IsStateTx(void)
{
  return m_state == TX;
}

bool
UanPhySimple::IsStateCcaBusy(void)
{
  return m_state == CCABUSY;
}


void
UanPhySimple::SetRxGainDb(double gain)
{
  m_rxGainDb = gain;

}
void
UanPhySimple::SetTxPowerDb(double txpwr)
{
  m_txPwrDb = txpwr;
}
void
UanPhySimple::SetRxThresholdDb(double thresh)
{
  m_rxThreshDb = thresh;
}
void
UanPhySimple::SetCcaThresholdDb(double thresh)
{
  m_ccaThreshDb = thresh;
}
double
UanPhySimple::GetRxGainDb(void)
{
  return m_rxGainDb;
}
double
UanPhySimple::GetTxPowerDb(void)
{
  return m_txPwrDb;

}
double
UanPhySimple::GetRxThresholdDb(void)
{
  return m_rxThreshDb;
}
double
UanPhySimple::GetCcaThresholdDb(void)
{
  return m_ccaThreshDb;
}

Ptr<UanChannel>
UanPhySimple::GetChannel(void) const
{
  return m_channel;
}

Ptr<UanNetDevice>
UanPhySimple::GetDevice(void)
{
  return m_device;
}

Ptr<UanTransducer>
UanPhySimple::GetTransducer(void)
{
  return m_transducer;
}
void
UanPhySimple::SetChannel(Ptr<UanChannel> channel)
{
  m_channel = channel;
}

void
UanPhySimple::SetDevice(Ptr<UanNetDevice> device)
{
  m_device = device;
}

void
UanPhySimple::SetMac(Ptr<UanMac> mac)
{
  m_mac = mac;
}

void
UanPhySimple::SetTransducer(Ptr<UanTransducer> trans)
{
  m_transducer = trans;
  m_transducer->AddPhy(this);
}


void
UanPhySimple::NotifyTransStartTx(Ptr<Packet> packet, double txPowerDb, UanTxMode mode)
{
  //TODO:  If receiving signal error
}

void
UanPhySimple::NotifyIntChange(void)
{
  if(m_state == CCABUSY && GetInterferenceDb((Ptr<Packet>) 0) < m_ccaThreshDb)
    {
      m_state = IDLE;
      NotifyListenersCcaEnd();
    }
}

double
UanPhySimple::CalculateSinrDb(Ptr<Packet> pkt, double rxPowerDb)
{

//  double interDb = GetInterferenceDb(pkt);
//  double noiseDb = m_channel->GetNoiseDbHz();
//
//  double niDb = KpToDb(DbToKp(interDb) + DbToKp(noiseDb));
//  double sinrDb = rxPowerDb + m_rxGainDb - niDb;

  return 1000;

}

double
UanPhySimple::GetInterferenceDb(Ptr<Packet> pkt)
{

  const UanTransducer::ArrivalList &arrivalList = m_transducer->GetArrivalList();

  UanTransducer::ArrivalList::const_iterator it = arrivalList.begin();

  double interfPower=0;

  for(;it != arrivalList.end();it++)
    {
      if(pkt != it->GetPacket())
        interfPower += DbToKp(it->GetRxPowerDb());
    }

  return KpToDb(interfPower);

}

double
UanPhySimple::DbToKp(double db)
{
  return pow(10, db/10.0);
}
double
UanPhySimple::KpToDb(double kp)
{
  return 10*log10(kp);
}

void
UanPhySimple::NotifyListenersRxStart(void)
{

}
void
UanPhySimple::NotifyListenersRxGood(void)
{

}
void
UanPhySimple::NotifyListenersRxBad(void)
{

}
void
UanPhySimple::NotifyListenersCcaStart(void)
{

}
void
UanPhySimple::NotifyListenersCcaEnd(void)
{

}

void
UanPhySimple::NotifyListenersTxStart(Time duration)
{

}

uint32_t
UanPhySimple::GetNModes(void)
{
  return m_modes.GetNModes();
}

UanTxMode
UanPhySimple::GetMode(uint32_t n)
{
  NS_ASSERT(n < m_modes.GetNModes());

  return m_modes[n];
}

Ptr<Packet>
UanPhySimple::GetPacketRx(void) const
{
  return m_pktRx;
}

} // namespace ns3
