/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

#ifndef UANPHY_H_
#define UANPHY_H_

#include "ns3/object.h"
#include "ns3/uan-mac.h"
#include "ns3/uan-tx-mode.h"
#include "ns3/uan-prop-model.h"
#include "ns3/uan-transducer.h"

namespace ns3
{


/**
 * \class UanPhyCalcSinr
 *
 * Class used for calculating SINR of packet in UanPhy.
 * Can be set to any derived class using attributes of UanPhy
 * to implement different models.
 */
class UanPhyCalcSinr : public ns3::Object
{

public:
  /**
   * \param pkt Packet to calculate SINR for
   * \param arrTime Arrival time of pkt
   * \param rxPowerDb The received signal strength of the packet in dB re 1 kPa
   * \param ambNoiseDb Ambient channel noise in dB re 1 kPa
   * \param mode TX Mode of pkt
   * \param pdp  Power delay profile of pkt
   * \param arrivalList  List of interfering arrivals given from Transducer.
   */
  virtual double CalcSinrDb(Ptr<Packet> pkt,
                         Time arrTime,
                         double rxPowerDb,
                         double ambNoiseDb,
                         UanTxMode mode,
                         UanPdp pdp,
                         const UanTransducer::ArrivalList &arrivalList
                         ) const =0;
  inline double DbToKp(double db) const {return pow(10, db/10.0);}
  inline double KpToDb(double kp) const {return 10*log10(kp);}
};

/**
 * \class UanPhyPer
 * Used in UanPhy for determining probability of packet error based on received SINR and
 * modulation (mode).  Can be set in UanPhy via attributes
 */
class UanPhyPer : public ns3::Object
{
public:
  /**
   * Calculates the prob. of packet error based on
   * SINR at the receiver and a tx mode.
   * \param sinrDb SINR at receiver
   * \param mode TX mode used to transmit packet
   * \returns Probability of packet error
   */
  virtual double CalcPer(double sinrDb, UanTxMode mode)=0;
};


/**
 * \class UanPhyListener
 *
 * \brief Interface for PHY event listener
 */
class UanPhyListener {
public:
  virtual void NotifyRxStart(void)=0;
  virtual void NotifyRxEndOk(void)=0;
  virtual void NotifyRxEndError(void)=0;
  virtual void NotifyCcaStart(void)=0;
  virtual void NotifyCcaEnd(void)=0;
  virtual void NotifyTxStart(Time duration)=0;
};

/**
 * \class UanPhy
 *
 * \brief Base class for UAN Phy models
 */
class UanPhy : public ns3::Object
{
public:

  enum State {IDLE, CCABUSY, RX, TX};

  /**
   * arg1: packet received successfully
   * arg2: snr of packet
   * arg3: mode of packet
   */
  typedef Callback<void, Ptr<Packet>, double, UanTxMode> RxOkCallback;

  /**
   * arg1: packet received successfully
   * arg2: snr of packet
   */
  typedef Callback<void, Ptr<Packet>, double > RxErrCallback;

  /**
   * \param pkt  Packet to transmit
   *
   */
  virtual void SendPacket(Ptr<Packet> pkt, uint32_t modeNum)=0;

  /**
   * \param listener New listener to register
   *
   * Register an object to be notified of common Phy events
   */
  virtual void RegisterListener(UanPhyListener *listener)=0;

  /**
   *
   * Packet arriving from channel:  i.e.  leading bit of packet has arrived.
   */
  virtual void StartRxPacket(Ptr<Packet> pkt, double rxPowerDb, UanTxMode txMode, UanPdp pdp)=0;

  /**
   *
   * \param cb Callback to be enabled when Packet is received without error
   */
  virtual void SetReceiveOkCallback (RxOkCallback cb)=0;

  /**
   *
   * \param cb Callback to be enabled when a packet is received with errors
   */
  virtual void SetReceiveErrorCallback (RxErrCallback cb)=0;

  /**
   * \param gain Gain added at receiver
   */

  virtual void SetRxGainDb(double gain)=0;

  /**
   * \param txpwr Final output transmission power in dB
   */
  virtual void SetTxPowerDb(double txpwr)=0;

  /**
   * \param thresh Threshold SINR for propper reception in dB
   */
  virtual void SetRxThresholdDb(double thresh)=0;

  /**
   * \param thresh Signal power at receiver required for CCA busy state
   *
   */
  virtual void SetCcaThresholdDb(double thresh)=0;

  /**
   * \returns Gain added to signal at receiver in dB
   */
  virtual double GetRxGainDb(void)=0;

  /**
   * \returns Current TX power output
   */
  virtual double GetTxPowerDb(void)=0;

  /**
   * \returns Required signal strength, in dB, to begin receiving packet
   */
  virtual double GetRxThresholdDb(void)=0;

  /**
   * \returns Threshold signal strength in dB to enter CCA busy mode
   */
  virtual double GetCcaThresholdDb(void)=0;

  /**
   *
   * \returns True if Phy is IDLE
   */
  virtual bool IsStateIdle(void)=0;
  /**
   *
   * \returns True if Phy is not IDLE
   */
  virtual bool IsStateBusy(void)=0;
  /**
   *
   * \returns True if Phy is currently in receive mode
   */
  virtual bool IsStateRx(void)=0;
  /**
   *
   * \returns True if Phy is busy transmitting
   */
  virtual bool IsStateTx(void)=0;
  /**
   *
   * \returns True if Phy is not TX or RX but the channel is sensed busy
   */
  virtual bool IsStateCcaBusy(void)=0;

  /**
   * \returns Channel this phy is attached to
   */
  virtual Ptr<UanChannel> GetChannel(void) const=0;

  /**
   * \returns the Net Device that this Phy is a part of
   */
  virtual Ptr<UanNetDevice> GetDevice(void)=0;

  /**
   * \param channel The channel this Phy is attached to
   */
  virtual void SetChannel(Ptr<UanChannel> channel)=0;

  /**
   * \param device The Net Device this Phy is a part of
   */
  virtual void SetDevice(Ptr<UanNetDevice> device)=0;

  /**
   * \param mac The MAC forwarding messages to this Phy
   */
  virtual void SetMac(Ptr<UanMac> mac)=0;

  /**
   * \param packet Packet that is beginning transmission
   * \param txPowerDb Transmit power of packet
   * \param txMode  Transmission mode of packet
   *
   * Notification that there is a transmission beginning
   * on the transducer that this Phy is attached to.
   */

  virtual void NotifyTransStartTx(Ptr<Packet> packet, double txPowerDb, UanTxMode txMode)=0;

  /**
   *
   * Function called when there has been a change in the
   * ammount of interference this node is experiencing
   * from other transmissions
   */
  virtual void NotifyIntChange(void)=0;

  /**
   *
   * \param trans Transducer this Phy outputs to / receives from
   */
  virtual void SetTransducer(Ptr<UanTransducer> trans)=0;

  /**
   * \returns Transducer this Phy outputs to / receives from
   */
  virtual Ptr<UanTransducer> GetTransducer(void)=0;

  /**
   * \returns Number of TX modes supported by this PHY
   */
  virtual uint32_t GetNModes(void)=0;

  /**
   * \param n Mode number of mode to return (place in Modeslist)
   * \returns Modes #n
   */
  virtual UanTxMode GetMode(uint32_t n)=0;

  /**
   * \warn Returns non-valid pointer if IsStateRx == false
   * \returns Packet currently being received in PHY
   */
  virtual Ptr<Packet> GetPacketRx(void) const=0;
};

}

#endif /*UANPHY_H_*/
