/*
 * uan-prop-model-bh.cc
 *
 *  Created on: 5-Nov-2008
 *      Author: ltracy
 */

#include "uan-prop-model-bh.h"
#include "uan-tx-mode.h"
#include "ns3/vector.h"
#include "ns3/mobility-model.h"
#include "ns3/log.h"
#include "ns3/string.h"

#include <fstream>
#include <sstream>
#include <iomanip>
#include <complex>
#include <cstdlib>
#include <algorithm>

using std::string;

NS_LOG_COMPONENT_DEFINE("UanPropModelBh");


namespace ns3
{

NS_OBJECT_ENSURE_REGISTERED(UanPropModelBh);

std::set<string>
BhConfig::FillParams(void)
{
  std::set<string> params;

  params.insert("datapath");
  params.insert("numsrcdepth");
  params.insert("minsrcdepth");
  params.insert("srcdepthstep");
  params.insert("numrecdepth");
  params.insert("minrecdepth");
  params.insert("recdepthstep");
  params.insert("numrange");
  params.insert("minrange");
  params.insert("rangestep");
  params.insert("minfreq");
  params.insert("numfreq");
  params.insert("freqstep");

  return params;
}

bool
BhConfig::Config(std::string cfgFileName)
{
  std::ifstream inf(cfgFileName.c_str());
  std::ostringstream buf;
  int (*tl)(int) = tolower;

  std::set<string> params = FillParams();

  if(!inf.good())
    {
      return false; //Bad environment file
    }
  m_ConfigFile = string(cfgFileName);

  string line;
  uint32_t lineNo=0;
  while(getline(inf, line))
    {
      lineNo++;
      trim(line);
      //NS_LOG_DEBUG("Line " << lineNo << ": " << line);
      if(line.empty() || line.at(0) == '#') //comment
         continue;

       uint32_t colonPos = line.find_first_of(':');
       if(colonPos == string::npos)
         {
           NS_LOG_WARN("WARNING: Unrecognized configuration line at line number " << lineNo << ": " << line);
           continue;
         }


       string configParam = trim(line.substr(0, colonPos));
       transform(configParam.begin(), configParam.end(), configParam.begin(), tl);


       uint32_t endPos = line.find_first_of("# ", colonPos+1);
       uint32_t valSize = (endPos == string::npos)?(line.length() - colonPos + 1):(colonPos - endPos - 1);
       string configValue = trim(line.substr(colonPos+1, valSize));


       if(configParam == "datapath")
         {
           m_DataPath = configValue;
           NS_LOG_DEBUG("Set data path: " << m_DataPath);

         }
       else if(configParam == "numsrcdepth")
         {
           m_NumSrcDepth = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set number of source depths: " << m_NumSrcDepth);
         }
       else if(configParam == "minsrcdepth")
         {
           m_MinSrcDepthM = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set minimum source depth: " << m_MinSrcDepthM << " m");
         }
       else if(configParam == "srcdepthstep")
         {
           m_SrcDepthStepM = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set source depth step interval: " << m_SrcDepthStepM << " m");
         }
       else if(configParam == "numrecdepth")
         {
           m_NumRecDepth = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set number of receiver depths: " << m_NumRecDepth << " depths");
         }
       else if(configParam == "minrecdepth")
         {
           m_MinRecDepthM = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set minimum receiver depth: " << m_MinRecDepthM << " m");
         }
       else if(configParam == "recdepthstep")
         {
           m_RecDepthStepM = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set receiver depth step: " << m_RecDepthStepM << " m");
         }
       else if(configParam == "numrange")
         {
           m_NumRanges = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set number of ranges: " << m_NumRanges);
         }
       else if(configParam == "minrange")
         {
           m_MinRangeKm = atof(configValue.c_str());
           NS_LOG_DEBUG("Set minimum range: " << m_MinRangeKm << " km");
         }
       else if(configParam == "rangestep")
         {
           m_RangeStepKm = atof(configValue.c_str());
           NS_LOG_DEBUG("Set range step interval: " << m_RangeStepKm << " km");
         }
       else if(configParam == "minfreq")
         {
           m_MinFreqHz = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set minimum frequency: " << m_MinFreqHz << " Hz");
         }
       else if(configParam == "numfreq")
         {
           m_NumFreqs = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set number of frequencies: " << m_NumFreqs);
         }
       else if(configParam == "freqstep")
         {
           m_FreqStepHz = atoi(configValue.c_str());
           NS_LOG_DEBUG("Set frequency step: " << m_FreqStepHz << " Hz");
         }

       else
         continue;
       params.erase(configParam);
    }

  inf.close();

  if(!params.empty())
    {
      NS_LOG_WARN("BhConfig:  Config file " << m_ConfigFile << " is incomplete");
      NS_LOG_WARN("Missing parameters:");
      std::set<string>::iterator it = params.begin();
      for(; it != params.end(); it++)
        {
          NS_LOG_WARN(*it);
        }

      return false;
    }
  m_Configured = true;
  return true;
}
BhIndex::BhIndex(uint32_t freqHz, uint32_t srcDepth, uint32_t recDepth, uint32_t rangeM) :
    m_FreqHz(freqHz),
    m_SrcDepthM(srcDepth),
    m_RecDepthM(recDepth),
    m_RangeM(rangeM)
{

}

BhIndex::BhIndex() :
    m_FreqHz(0),
    m_SrcDepthM(0),
    m_RecDepthM(0),
    m_RangeM(0)
{

}
bool
BhIndex::operator==(const BhIndex &other) const
{
  return m_FreqHz == other.m_FreqHz && m_SrcDepthM == other.m_SrcDepthM
         && m_RecDepthM == other.m_RecDepthM
         && m_RangeM == other.m_RangeM;
}
bool
BhIndex::operator < (const BhIndex &other) const
{
  if(m_FreqHz != other.m_FreqHz)
    return m_FreqHz < other.m_FreqHz;
  if(m_SrcDepthM != other.m_SrcDepthM)
    return m_SrcDepthM < other.m_SrcDepthM;
  if(m_RecDepthM != other.m_RecDepthM)
    return m_RecDepthM < other.m_RecDepthM;
  return m_RangeM < other.m_RangeM;
}
bool
BhIndex::operator > (const BhIndex &other) const
{
  return !(*this == other || *this < other);
}

BellhopResp::BellhopResp(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM, UanPdp pdp, double pathLossDb)
   : valid(true),
     m_Index(freqHz, srcDepthM, recDepthM, rangeM),
     m_Pdp(pdp),
     m_PathLossDb(pathLossDb)
{
}

BellhopResp::BellhopResp()
  : valid(false)
{
}

BhRespManager &
BhRespManager::GetInstance()
{
  static BhRespManager manager;
  return manager;
}

BhIndex
BhRespManager::MakeIndex(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM)
{

  uint32_t approxFreq;
  if(freqHz < m_Config.GetMinFreqHz())
    {
      NS_LOG_WARN("Frequency lookup for " << freqHz << "Hz is less than minimum frequency");
      approxFreq = m_Config.GetMinFreqHz();
    }

  else if(freqHz > m_Config.GetMaxFreqHz())
    {
      NS_LOG_WARN("Frequency lookup for " << freqHz << "Hz is greather than maximum frequency");
      approxFreq = m_Config.GetMaxFreqHz();
    }
  else
    {
      double freqTemp = freqHz - m_Config.GetMinFreqHz();
      freqTemp = freqTemp / m_Config.GetFreqStepHz() + 0.5;
      approxFreq = m_Config.GetMinFreqHz() + ( (uint32_t) freqTemp)*m_Config.GetFreqStepHz();
    }

  uint32_t approxSrcDepth;
  if(srcDepthM > m_Config.GetMaxSrcDepthM())
    {
      NS_LOG_WARN("Source depth lookup for " << srcDepthM << "m is greather than maximum depth");
      approxSrcDepth = m_Config.GetMaxSrcDepthM();
    }
  else if(srcDepthM < m_Config.GetMinSrcDepthM())
    {
      NS_LOG_WARN("Source depth lookup for " << srcDepthM << "m is less than minimum depth");
      approxSrcDepth = m_Config.GetMinSrcDepthM();
    }
  else
    {
      double sdTmp = srcDepthM - m_Config.GetMinSrcDepthM();
      sdTmp = sdTmp/m_Config.GetSrcDepthStepM() + 0.5;
      approxSrcDepth = ((uint32_t) sdTmp) *  m_Config.GetSrcDepthStepM() + m_Config.GetMinSrcDepthM();
    }



  uint32_t approxRecDepth;

  if(recDepthM > m_Config.GetMaxRecDepthM())
    {
      NS_LOG_WARN("Receiver depth lookup for " << recDepthM << "m is greater than maximum depth");
      approxRecDepth = m_Config.GetMaxRecDepthM();
    }
  else if(recDepthM < m_Config.GetMinRecDepthM())
    {
      NS_LOG_WARN("Receiver depth lookup for " << recDepthM << "m is less than minimum depth");
      approxRecDepth = m_Config.GetMinRecDepthM();
    }
  else
    {
      double rdTmp = recDepthM - m_Config.GetMinRecDepthM();
      rdTmp = rdTmp/m_Config.GetRecDepthStepM() + 0.5;
      approxRecDepth = ((uint32_t) rdTmp) *  m_Config.GetRecDepthStepM() + m_Config.GetMinRecDepthM();
    }

  uint32_t approxRange;

  if(rangeM > m_Config.GetMaxRangeM())
    {
      NS_LOG_WARN("Range lookup for " << rangeM << "m is greater than maximum range");
      approxRange = m_Config.GetMaxRangeM();
    }
  else if(rangeM < m_Config.GetMinRangeM())
    {
      NS_LOG_WARN("Range lookup for " << rangeM << "m is less than minimum range");
      approxRange = m_Config.GetMinRangeM();
    }
  else
    {
      double rrTmp = rangeM - m_Config.GetMinRangeM();
      rrTmp = rrTmp/m_Config.GetRangeStepM() + 0.5;
      approxRange = ((uint32_t) rrTmp) *  m_Config.GetRangeStepM() + m_Config.GetMinRangeM();

    }
  //NS_LOG_DEBUG("Found approx range=" << approxRange << "   From range=" << rangeM);

  return ns3::BhIndex(approxFreq, approxSrcDepth, approxRecDepth, approxRange);

}
bool
BhRespManager::IsConfigured()
{
  return m_Config.IsConfigured();
}

bool
BhRespManager::Configure(std::string configFileName)
{

  return m_Config.Config(configFileName);
}

UanPdp
BhRespManager::GetPdp(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM)
{
  BellhopResp &resp = GetResp(freqHz, srcDepthM, recDepthM, rangeM);
  return resp.GetPdp();
}

double
BhRespManager::GetPathlossDb(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM)
{
  BellhopResp &resp = GetResp(freqHz, srcDepthM, recDepthM, rangeM);
  return resp.GetPathLossDb();
}

BellhopResp &
BhRespManager::GetResp(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM)
{
  NS_ASSERT(IsConfigured());

  BhIndex index = MakeIndex(freqHz, srcDepthM, recDepthM, rangeM);

  if(m_BhData.find(index) == m_BhData.end())
    {
      NS_LOG_DEBUG("Getting from file: sd=" << index.m_SrcDepthM << " rd=" << index.m_RecDepthM << " rr=" << index.m_RangeM << " freq=" << index.m_FreqHz);
      RetreiveFromFile(index);
    }



  return m_BhData[index];

}

void
BhRespManager::RetreiveFromFile(BhIndex index)
{
  std::ostringstream buf;
  std::string lineBuf;

  buf << std::fixed << std::setprecision(1) << m_Config.GetDataPath() << index.m_SrcDepthM << "-" << index.m_FreqHz << ".dat";

  std::ifstream inf(buf.str().c_str());
  if(!inf.is_open())
    {
      NS_FATAL_ERROR("UanPropBh:  Could not open data file, " << buf.str());
    }

  buf.str("");
  buf << "rd " << index.m_RecDepthM;
  while(std::getline(inf, lineBuf))
    {
      if(lineBuf == buf.str())
        break;
    }
  if(!inf.good())
    {
      NS_FATAL_ERROR("UanPropBh: Could not find receiver depth " << index.m_RecDepthM
                  << " for source depth " << index.m_SrcDepthM
                  << " and freq " << index.m_FreqHz);
    }

  buf.str("");
  buf << "rr " << index.m_RangeM;
  while(std::getline(inf, lineBuf))
    {
      if(lineBuf == buf.str())
        break;
    }
  if(!inf.good())
    {
      NS_LOG_DEBUG("Looking for range string: " << buf.str());
      NS_FATAL_ERROR("UanPropBh: Could not find range " << index.m_RangeM
                  << " for source depth " << index.m_SrcDepthM
                  << " and freq " << index.m_FreqHz);
    }
  std::getline(inf, lineBuf);
  double pathLossDb = atof(lineBuf.c_str());

  UanPdp pdp;
  inf >> pdp;

  //TODO:
  //Normalize PDP and calculate pathloss.
  //This should probably be done at time of data file
  //creation
  //or maybe moved somewhere more appropriate
  UanPdp::Iterator pdpIt = pdp.GetBegin();
  //double pLossDb = 0;
  double pLoss = 0;
  Time minDelay = Seconds(1e30);
  for(;pdpIt != pdp.GetEnd(); pdpIt++)
    {
      if((*pdpIt).GetDelay() < minDelay)
        minDelay = (*pdpIt).GetDelay();
      NS_LOG_DEBUG("Reading amplitude" << (*pdpIt).GetAmp());
      pLoss += abs((*pdpIt).GetAmp());
    }
  NS_LOG_DEBUG("Total sum alphas: " << pLoss);
  NS_LOG_DEBUG("Pathloss: " << pathLossDb);

//  pLossDb += 20*log10(pLoss);
//  NS_LOG_DEBUG("Got pathlossDb = " << pLossDb);

  UanPdp normPdp;

//  for(pdpIt = pdp.GetBegin();pdpIt != pdp.GetEnd(); pdpIt++)
//    {
//      normPdp.AddTap(Tap((*pdpIt).GetDelay() - minDelay, (*pdpIt).GetAmp()/pLoss));
//    }

  BellhopResp resp(index.m_FreqHz, index.m_SrcDepthM, index.m_RecDepthM, index.m_RangeM, pdp, pathLossDb);

  m_BhData.insert(std::make_pair(index, resp));
}

UanPropModelBh::UanPropModelBh()
{
  // TODO Auto-generated constructor stub

}

UanPropModelBh::~UanPropModelBh()
{
  // TODO Auto-generated destructor stub
}

TypeId
UanPropModelBh::GetTypeId(void)
{
  static TypeId tid = TypeId("ns3::UanPropModelBh")
    .SetParent<Object>()
    .AddConstructor<UanPropModelBh>()
    .AddAttribute("ConfigFile","Path and filename of configuration file",
                  StringValue("dat/defaultbh.cfg"),
                  MakeStringAccessor(&UanPropModelBh::m_CfgFileName),
                  MakeStringChecker());
  return tid;
}

double
UanPropModelBh::GetPathLossDb(Ptr<MobilityModel> src, Ptr<MobilityModel> dest, UanTxMode mode)
{
  uint32_t freqHz = mode.GetCenterFreqHz();
  uint32_t srcDepth = (uint32_t) (src->GetPosition().z);
  uint32_t recDepth = (uint32_t) (dest->GetPosition().z);

  BhRespManager &manager = BhRespManager::GetInstance();

  if(!manager.IsConfigured())
    {
      if(!manager.Configure(m_CfgFileName))
        {
          NS_FATAL_ERROR("UanPropModelBh:  Could not open config file " << m_CfgFileName);
        }
    }


  Vector srcPosition = src->GetPosition();
  Vector destPosition = dest->GetPosition();
  double dx = srcPosition.x - destPosition.x;
  double dy = srcPosition.y - destPosition.y;

  uint32_t range = (uint32_t) (sqrt(dx*dx + dy*dy)+0.5);
  return manager.GetPathlossDb(freqHz, srcDepth, recDepth, range);

}
UanPdp
UanPropModelBh::GetPdp(Ptr<MobilityModel> src, Ptr<MobilityModel> dest, UanTxMode mode)
{
  uint32_t freqHz = mode.GetCenterFreqHz();
  uint32_t srcDepth = (uint32_t) (src->GetPosition().z);
  uint32_t recDepth = (uint32_t) (dest->GetPosition().z);

  BhRespManager &manager = BhRespManager::GetInstance();
  if(!manager.IsConfigured())
    {
      if(!manager.Configure(m_CfgFileName))
        {
          NS_FATAL_ERROR("UanPropModelBh:  Could not open config file " << m_CfgFileName);
        }
    }

  Vector srcPosition = src->GetPosition();
  Vector destPosition = dest->GetPosition();
  double dx = srcPosition.x - destPosition.x;
  double dy = srcPosition.y - destPosition.y;

  uint32_t range = (uint32_t) (sqrt(dx*dx + dy*dy)+0.5);
  return manager.GetPdp(freqHz, srcDepth, recDepth, range);
}
Time
UanPropModelBh::GetDelay(Ptr<MobilityModel> a, Ptr<MobilityModel> b, UanTxMode mode)
{
  return Seconds(a->GetDistanceFrom(b)/1500.0);
}

string& BhConfig::trimleft( string& s )
{
   string::iterator it;

   for( it = s.begin(); it != s.end(); it++ )
      if( !isspace( *it ) )
         break;

   s.erase( s.begin(), it );
   return s;
}

string& BhConfig::trimright( string& s )
{
   string::difference_type dt;
   string::reverse_iterator it;

   for( it = s.rbegin(); it != s.rend(); it++ )
      if( !isspace( *it ) )
         break;

   dt = s.rend() - it;

   s.erase( s.begin() + dt, s.end() );
   return s;
}

string& BhConfig::trim( string& s )
{
   trimleft( s );
   trimright( s );
   return s;
}

string BhConfig::trim( const string& s )
{
   string t = s;
   return trim( t );
}

}
