/*
 * uan-prop-model-bh.h
 *
 *  Created on: 5-Nov-2008
 *      Author: ltracy
 */

#ifndef UANPROPMODELBH_H_
#define UANPROPMODELBH_H_

#include "uan-prop-model.h"

#include <string>
#include <map>
#include <set>
#include <set>

using std::string;

namespace ns3
{

/**
 * \class BhConfig
 *
 * \brief Holds configuration information for Bellhop environment
 */
class BhConfig
{
public:
  BhConfig() : m_Configured(false) { }
  virtual ~BhConfig() { }
  virtual bool Config(std::string cfgFile);

  bool IsConfigured(void) {return m_Configured;}
  std::string GetDataPath() { return m_DataPath; }
  std::string GetCfgFileName() { return m_ConfigFile; }
  uint32_t GetMinSrcDepthM() { return m_MinSrcDepthM; }
  uint32_t GetMaxSrcDepthM() { return m_MinSrcDepthM + (m_NumSrcDepth-1)*m_SrcDepthStepM; }
  uint32_t GetSrcDepthStepM() { return m_SrcDepthStepM; }

  uint32_t GetMinRecDepthM() { return m_MinRecDepthM; }
  uint32_t GetMaxRecDepthM() { return m_MinRecDepthM + (m_NumRecDepth-1)*m_RecDepthStepM; }
  uint32_t GetRecDepthStepM() { return m_RecDepthStepM; }

  uint32_t GetMinRangeM() { return (uint32_t) (m_MinRangeKm*1000.0); }
  uint32_t GetMaxRangeM() { return (uint32_t) (1000.0*(m_MinRangeKm + (m_NumRanges-1)*m_RangeStepKm)); }
  uint32_t GetRangeStepM() { return (uint32_t) (m_RangeStepKm*1000.0); }

  uint32_t GetMinFreqHz() { return m_MinFreqHz; }
  uint32_t GetMaxFreqHz() { return m_MinFreqHz + (m_NumFreqs-1)*m_FreqStepHz;}
  uint32_t GetFreqStepHz() { return m_FreqStepHz; }

  static string& trimleft( string& s );
  static string& trimright( string& s );
  static string& trim( string& s );
  static string trim( const string& s );

protected:
  bool m_Configured;

  string m_DataPath;
  string m_ConfigFile;

  uint32_t m_NumSrcDepth;
  uint32_t m_MinSrcDepthM;
  uint32_t m_SrcDepthStepM;

  uint32_t m_NumRecDepth;
  uint32_t m_MinRecDepthM;
  uint32_t m_RecDepthStepM;

  uint32_t m_NumRanges;
  double m_MinRangeKm;
  double m_RangeStepKm;

  uint32_t m_NumFreqs;
  uint32_t m_MinFreqHz;
  uint32_t m_FreqStepHz;

private:
  std::set<string> FillParams(void);
};

struct BhIndex
{

  BhIndex(uint32_t freqHz, uint32_t srcDepth, uint32_t recDepth, uint32_t rangeM);
  BhIndex();

  uint32_t m_FreqHz;
  uint32_t m_SrcDepthM;
  uint32_t m_RecDepthM;
  uint32_t m_RangeM;

  bool operator ==(const BhIndex &other) const;
  bool operator <(const BhIndex &other) const;
  bool operator >(const BhIndex &other) const;
};

class BellhopResp
{
public:

  BellhopResp(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM, UanPdp pdp, double pathLossDb);
  BellhopResp();


  uint32_t GetFreqHz(void) const { return m_Index.m_FreqHz; }
  uint32_t GetSrcDepthM(void) const { return m_Index.m_SrcDepthM; }
  uint32_t GetRecDepthM(void) const { return m_Index.m_RecDepthM; }
  uint32_t GetRangeM(void) const { return m_Index.m_RangeM; }
  UanPdp GetPdp(void) const { return m_Pdp; }
  double GetPathLossDb(void) const { return m_PathLossDb; }
  const BhIndex &GetIndex(void) {return m_Index;}

private:
  bool valid;
  BhIndex m_Index;
  UanPdp m_Pdp;
  double m_PathLossDb;
};

/**
 * \class BHRespManager
 * \brief This class manages impulse responses received from Bellhop
 *
 * Returns a reference to object containing list of arrival information
 * There will only be a single instance of this class.
 */
class BhRespManager {
public:
  static BhRespManager &GetInstance();


  BhRespManager()  {}
  virtual ~BhRespManager() { }



  UanPdp GetPdp(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM);
  double GetPathlossDb(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM);
  /*
   * \returns True if successfully configured
   */
  bool IsConfigured();

  /*
   * \param configFileName Full name and path to config file
   */
  bool Configure(std::string configFileName);


  BhIndex MakeIndex(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM);

private:
  ///Updates information in dat;
  //void updateDat(vector<int> v);
  ///Gets Bellhop information from file.

  //virtual void lookup(BellhopResp &res, double freq, int srcDepth, int recDepth, int range);


  std::map<BhIndex, BellhopResp> m_BhData;
  BhConfig m_Config;
  void RetreiveFromFile(BhIndex index);
  BellhopResp &GetResp(uint32_t freqHz, uint32_t srcDepthM, uint32_t recDepthM, uint32_t rangeM);


};

/**
 * \class UanPropModelBh
 * \brief Returns stored Bellhop environment information for use in determining prop. information
 *
 * This class makes use of stored Bellhop runs
 */
class UanPropModelBh: public ns3::UanPropagationModel
{
public:
  UanPropModelBh();
  virtual ~UanPropModelBh();

  static TypeId GetTypeId(void);

  //Inherited methods
  virtual double GetPathLossDb(Ptr<MobilityModel> a, Ptr<MobilityModel> b, UanTxMode mode);
  virtual UanPdp GetPdp(Ptr<MobilityModel> src, Ptr<MobilityModel> dest, UanTxMode mode);
  virtual Time GetDelay(Ptr<MobilityModel> src, Ptr<MobilityModel> dest, UanTxMode mode);

  void SetConfigFile(std::string fileName);
  std::string GetConfigFile();

private:

  std::string m_CfgFileName;
  BhConfig m_Config;
};

}

#endif /* UANPROPMODELBH_H_ */
