/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

#include "uan-prop-model.h"
#include "ns3/nstime.h"
#include <complex>
#include <set>

namespace ns3
{

std::ostream &
operator<<(std::ostream &os, UanPdp &pdp)
{
  os << pdp.GetNTaps() << '|';
  std::set<Tap>::iterator it = pdp.m_taps.begin();
  for(;it!= pdp.m_taps.end();it++)
    {
      os << (*it).GetDelay().GetSeconds() << '|' << (*it).GetAmp() << '|';
    }
  return os;

}
std::istream &
operator>>(std::istream &is, UanPdp &pdp)
{
  uint32_t ntaps;
  char c1, c2;

  is >> ntaps >> c1;
  if(c1 != '|')
    {
      NS_FATAL_ERROR("UanPdp data corrupted");
      return is;
    }

  double delay;
  std::complex<double> amp;

  for(uint32_t i=0;i<ntaps;i++)
    {
      is >> delay >> c1 >> amp >> c2;
      if(!(c1 == '|' && c2 == '|'))
        {
          NS_FATAL_ERROR("UanPdp data corrupted");
          return is;
        }
      pdp.m_taps.insert(Tap(Seconds(delay), amp));
    }
  return is;

}

Tap::Tap(Time delay, std::complex<double> amp) :
  m_amplitude(amp),
  m_delay(delay)
{

}

std::complex<double>
Tap::GetAmp(void) const
{
  return m_amplitude;
}

Time
Tap::GetDelay(void) const
{
  return m_delay;
}

bool
Tap::operator<( const Tap &other) const
{
  return m_delay < other.m_delay;
}

bool
Tap::operator<=(const Tap &other) const
{
  return m_delay <= other.m_delay;
}

bool
Tap::operator>(const Tap &other) const
{
  return m_delay > other.m_delay;
}

bool
Tap::operator>=(const Tap &other) const
{
  return m_delay >= other.m_delay;
}

bool
Tap::operator==(const Tap &other) const
{
  return m_delay == other.m_delay;
}

bool
Tap::operator!=(const Tap &other) const
{
  return m_delay != other.m_delay;
}

UanPdp::UanPdp()
{

}

UanPdp::UanPdp(std::vector<Tap> taps)
{
  m_taps.insert(taps.begin(), taps.end());
}

void
UanPdp::AddTap(Tap tap)
{
  m_taps.insert(tap);
}

UanPdp::Iterator
UanPdp::GetBegin(void)
{
  return m_taps.begin();
}

UanPdp::Iterator
UanPdp::GetEnd(void)
{
  return m_taps.end();
}

uint32_t
UanPdp::GetNTaps(void)
{
  return m_taps.size();
}

double
UanPdp::SumTapsNc(Time begin, Time end)
{
  Iterator it = GetBegin();
  Time pdelay = it->GetDelay();
  double sum=0;
  for(;it != GetEnd();it++)
    {
      if(it->GetDelay() - pdelay > end)
        {
          break;
        }
      else if(it->GetDelay() - pdelay >= begin)
        {
          sum += abs(it->GetAmp());
        }
    }
  return sum;
}

std::complex<double>
UanPdp::SumTapsC(Time begin, Time end)
{
  Iterator it = GetBegin();
  Time pdelay = it->GetDelay();
  std::complex<double> sum=0;
  for(;it != GetEnd();it++)
    {
      if(it->GetDelay() - pdelay > end)
        {
          break;
        }
      else if(it->GetDelay() - pdelay >= begin)
        {
          sum += it->GetAmp();
        }
    }
  return sum;
}

UanPdp
UanPdp::CreateImpulsePdp(void)
{
  UanPdp pdp;
  pdp.AddTap(Tap(Seconds(0), 1));
  return pdp;
}

}
