/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */

#ifndef UANPROPMODEL_H_
#define UANPROPMODEL_H_

#include "ns3/object.h"
#include "ns3/mobility-model.h"
#include "ns3/nstime.h"


#include <vector>
#include <complex>
#include <utility>
#include <set>
namespace ns3
{

class UanTxMode;

/**
 * \class Tap
 * \brief Holds PDP arrival information
 */
class Tap {
public:
  Tap() : m_amplitude(0), m_delay(Seconds(0)) { }
  /**
   * \param delay Time delay (usually from first arrival) of signal
   * \param amp Complex amplitude of arrival
   */
  Tap(Time delay, std::complex<double> amp);
  /**
   * \returns Complex amplitude of arrival
   */
  std::complex<double> GetAmp(void) const;
  /**
   * \returns Time delay (usually from first arrival) of signal
   */
  Time GetDelay(void) const;

  bool operator<(const Tap &other) const;
  bool operator<=(const Tap &other) const;
  bool operator>(const Tap &other) const;
  bool operator>=(const Tap &other) const;
  bool operator==(const Tap &other) const;
  bool operator!=(const Tap &other) const;

private:
  std::complex<double> m_amplitude;
  Time m_delay;
};

/**
 * \class UanPdp
 *
 * Container class to describe power delay profile returned
 * from UAN propagation models.
 *
 * Generally, the profile should be normalized, such that
 * the sum of all taps should equal 1.  The received signal
 * power on any interval (t1, t2) can then be found from
 * summing the taps on the interval and multiplying by
 * the total received power at the receiver.
 *
 *
 */
class UanPdp {
public:
  typedef std::set<Tap>::const_iterator Iterator;
  UanPdp();
  UanPdp(std::vector<Tap> taps);
  void AddTap(Tap tap);
  Iterator GetBegin(void);
  Iterator GetEnd(void);
  uint32_t GetNTaps(void);
  double SumTapsNc(Time begin, Time end);
  std::complex<double> SumTapsC(Time begin, Time end);

  static UanPdp CreateImpulsePdp(void);
private:
  friend std::ostream &operator<<(std::ostream &os, UanPdp &pdp);
  friend std::istream &operator>>(std::istream &is, UanPdp &pdp);
  std::set<Tap> m_taps;
};
std::ostream &operator<<(std::ostream &os, UanPdp &pdp);
std::istream &operator>>(std::ostream &is, UanPdp &pdp);

/**
 * \class UanPropagationModel
 *
 * Base class for implemented underwater propagation models
 */
class UanPropagationModel : public ns3::Object
{
public:


  /**
   * Computes pathloss between nodes a and b.
   * \returns Pathloss in dB re 1 kPa
   */
  virtual double GetPathLossDb(Ptr<MobilityModel> a, Ptr<MobilityModel> b, UanTxMode txMode)=0;

  /**
   * \returns PDP for link between nodes a and b
   */
  virtual UanPdp GetPdp(Ptr<MobilityModel> a, Ptr<MobilityModel> b, UanTxMode mode)=0;
  /**
   * Finds propagation delay between nodes a and b
   * \returns Propagation delay
   */
  virtual Time GetDelay(Ptr<MobilityModel> a, Ptr<MobilityModel> b, UanTxMode mode)=0;

};

}

#endif /*UANPROPMODEL_H_*/
