/*
 * Copyright (c) 2008 University of Washington
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Leonard Tracy <lentracy@u.washington.edu>
 *
 *
 */


#include "uan-transducer-hd.h"
#include "ns3/simulator.h"
#include "ns3/uan-prop-model.h"
#include "uan-phy.h"
#include "uan-channel.h"
#include "ns3/log.h"
#include "ns3/pointer.h"

NS_LOG_COMPONENT_DEFINE("UanTransducerHd");

namespace ns3
{

NS_OBJECT_ENSURE_REGISTERED(UanTransducerHd);
UanTransducerHd::UanTransducerHd() :
  m_state(RX),
  m_endTxTime(Seconds(0))
{
}

UanTransducerHd::~UanTransducerHd()
{
}

TypeId
UanTransducerHd::GetTypeId()
{
  static TypeId tid = TypeId("ns3::UanTransducerHd")
    .SetParent<Object> ()
    .AddConstructor<UanTransducerHd>()
    ;
  return tid;
}

UanTransducer::State
UanTransducerHd::GetState() const
{
  return m_state;
}


bool
UanTransducerHd::IsRx(void) const
{
  return m_state==RX;
}

bool
UanTransducerHd::IsTx(void) const
{
  return m_state==TX;

}

const UanTransducer::ArrivalList &
UanTransducerHd::GetArrivalList(void) const
{
  return m_arrivalList;
}

void
UanTransducerHd::Receive(Ptr<Packet> packet, double rxPowerDb, UanTxMode txMode, UanPdp pdp)
{
  UanPacketArrival arrival(packet, rxPowerDb, txMode, pdp, Simulator::Now());

  m_arrivalList.push_back(arrival);
  Time txDelay = Seconds(packet->GetSize()*8.0/txMode.GetDataRateBps());
  Simulator::Schedule(txDelay, &UanTransducerHd::RemoveArrival, this, arrival);
  NS_LOG_DEBUG("" << Simulator::Now().GetSeconds() << " Transducer in receive");
  if(m_state == RX)
    {
      NS_LOG_DEBUG("Transducer state = RX");
      for(UanPhyList::const_iterator it=m_phyList.begin();it != m_phyList.end();it++)
        {
          NS_LOG_DEBUG("Calling StartRx");
          (*it)->StartRxPacket(packet, rxPowerDb, txMode, pdp);
        }
    }
}

void
UanTransducerHd::Transmit(Ptr<UanPhy> src, Ptr<Packet> packet, double txPowerDb, UanTxMode txMode)
{

  Time endTxTime;
  if(m_state == TX)
    {
      Simulator::Remove(m_endTxEvent);
    }
  else
    {
      m_state = TX;
    }


  Time delay = Seconds(packet->GetSize()*8.0/txMode.GetDataRateBps());
  NS_LOG_DEBUG("Transducer transmitting:  TX delay = " << delay << " seconds for packet size " << packet->GetSize() << " bytes and rate = " << txMode.GetDataRateBps() << " bps");
  for(UanPhyList::const_iterator it=m_phyList.begin();it != m_phyList.end();it++)
    {
      if(src != (*it))
        (*it)->NotifyTransStartTx(packet, txPowerDb, txMode);
    }
  m_channel->TxPacket(Ptr<UanTransducer> (this), packet, txPowerDb, txMode);


  delay = std::max(delay, m_endTxTime - Simulator::Now());

  m_endTxEvent = Simulator::Schedule(delay, &UanTransducerHd::EndTx, this);
  m_endTxTime = Simulator::Now() + delay;
}

void
UanTransducerHd::EndTx(void)
{
  NS_ASSERT(m_state==TX);
  //NS_LOG_DEBUG(Simulator::Now().GetSeconds() << " Transducer moving to state RX");
  m_state = RX;
  m_endTxTime = Seconds(0);
}
void
UanTransducerHd::SetChannel(Ptr<UanChannel> chan)
{
  NS_LOG_DEBUG("Transducer setting channel");
  m_channel = chan;

}
Ptr<UanChannel>
UanTransducerHd::GetChannel(void) const
{
  return m_channel;
}
void
UanTransducerHd::AddPhy(Ptr<UanPhy> phy)
{
  m_phyList.push_back(phy);
}

const UanTransducer::UanPhyList &
UanTransducerHd::GetPhyList(void) const
{
  return m_phyList;
}

void
UanTransducerHd::RemoveArrival(UanPacketArrival arrival)
{

  //Remove entry from arrival list
  for(ArrivalList::iterator it=m_arrivalList.begin();it != m_arrivalList.end();it++)
    {
      if(it->GetPacket() == arrival.GetPacket())
        {
          m_arrivalList.erase(it);
          break;
        }
    }

  for(UanPhyList::const_iterator it=m_phyList.begin();it != m_phyList.end();it++)
    {
      (*it)->NotifyIntChange();
    }

}

} // namespace ns3
